
static ERROR ResizeLayer(struct Render *, LONG, LONG, LONG, LONG, LONG, LONG, LONG, DOUBLE, LONG);

/****************************************************************************

###FIELD###
Name:   AbsX
Short:  The absolute horizontal position of a render object.
Type:   LONG
Status: Get/Set

###DESCRIPTION###
<p>If you need to know the absolute horiziontal position of a render object
then you can read this field.  The absolute value is calculated based on the
render object's position relative to the top most render object in the local
hierarchy.</p>

<p>It is possible to set this field, but only after initialisation of the
render object has occurred.</p>

###SEE ALSO###
Field: AbsY

###END###

****************************************************************************/

static ERROR GET_AbsX(struct Render *Self, LONG *Value)
{
   struct RenderList *list;
   LONG i;

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      for (i=0; list[i].ObjectID; i++) {
         if (list[i].ObjectID IS Self->Head.UniqueID) {
            *Value = list[i].AbsX;
            ReleaseRenderList(ARF_READ);
            return(ERR_Okay);
         }
      }

      ReleaseRenderList(ARF_READ);
      return(ObjectError(Self, ERH_Function, ERR_Search));
   }
   else return(ObjectError(Self, ERH_Function, ERR_AccessMemory));
}

static ERROR SET_AbsX(struct Render *Self, LONG *Value)
{
   struct RenderList *list;
   LONG i, parent;

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      for (i=0; list[i].ObjectID; i++) {
         if (list[i].ObjectID IS Self->Head.UniqueID) {
            list[i].AbsX = *Value;
            for (parent=i-1; (parent > 0) AND (list[parent].ObjectID != list[i].ParentID); parent--);
            list[i].XCoord = list[i].AbsX - list[parent].AbsX;
            Self->XCoord = list[i].XCoord;

            ReleaseRenderList(ARF_READ);
            return(ERR_Okay);
         }
      }

      ReleaseRenderList(ARF_READ);
      return(ObjectError(Self, ERH_Function, ERR_Search));
   }
   else return(ObjectError(Self, ERH_Function, ERR_AccessMemory));
}

/****************************************************************************

###FIELD###
Name:   AbsY
Short:  The absolute vertical position of a render object.
Type:   LONG
Status: Get/Set

###DESCRIPTION###
<p>If you need to know the absolute vertical position of a render object
then you can read this field.  The absolute value is calculated based on the
render object's position relative to the top most render object in the local
hierarchy.</p>

<p>It is possible to set this field, but only after initialisation of the
render object has occurred.</p>

###SEE ALSO###
Field: AbsX

###END###

****************************************************************************/

static ERROR GET_AbsY(struct Render *Self, LONG *Value)
{
   struct RenderList *list;
   LONG i;

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      for (i=0; list[i].ObjectID; i++) {
         if (list[i].ObjectID IS Self->Head.UniqueID) {
            *Value = list[i].AbsY;
            ReleaseRenderList(ARF_READ);
            return(ERR_Okay);
         }
      }

      ReleaseRenderList(ARF_READ);
      return(ObjectError(Self, ERH_Function, ERR_Search));
   }
   else return(ObjectError(Self, ERH_Function, ERR_AccessMemory));
}

static ERROR SET_AbsY(struct Render *Self, LONG *Value)
{
   struct RenderList *list;
   LONG i, parent;

   if ((list = AccessRenderList(ARF_READ, NULL))) {
      for (i=0; list[i].ObjectID; i++) {
         if (list[i].ObjectID IS Self->Head.UniqueID) {
            list[i].AbsY = *Value;
            for (parent=i-1; (parent > 0) AND (list[parent].ObjectID != list[i].ParentID); parent--);
            list[i].YCoord = list[i].AbsY - list[parent].AbsY;
            Self->YCoord = list[i].YCoord;

            ReleaseRenderList(ARF_READ);
            return(ERR_Okay);
         }
      }

      ReleaseRenderList(ARF_READ);
      return(ObjectError(Self, ERH_Function, ERR_Search));
   }
   else return(ObjectError(Self, ERH_Function, ERR_AccessMemory));
}

/*****************************************************************************

###FIELD###
Name:   Align
Short:  This field allows you to align a rendered area within its owner.
Type:   LONG/FLAGS
Prefix: ALIGN
Status: Read/Write

###DESCRIPTION###
<p>If you would like to set an abstract position for a drawable area, you can
give it an alignment.  This feature is most commonly used for horizontal and
vertical centering, as aligning to the the edges of a rendered area is
already handled by existing dimension fields.  Note that setting the alignment
over-rides any settings in related coordinate fields.  Valid alignment flags
are as follows:</p>

<blockquote>
<p>BOTTOM<br>
CENTER/MIDDLE<br>
LEFT<br>
HORIZONTAL<br>
RIGHT<br>
TOP<br>
VERTICAL</p>
</blockquote>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Bottom
Short:  Returns the bottom-most coordinate of a render object.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>You can read this field to find out what the bottom-most coordinate of
a render object is.  This is basically the opposite of the YCoord field,
and is calculated by adding the YCoord and Height fields together.</p>

###SEE ALSO###
Field: Right, YCoord

###END###

*****************************************************************************/

static ERROR GET_Bottom(struct Render *Self, LONG *Bottom)
{
   *Bottom = Self->YCoord + Self->Height;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   BottomLimit
Short:  Prevents a render object from moving beyond a given point at the bottom of its container.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>You can prevent a render object from moving beyond a given point at the
bottom of its container by setting this field.  If for example you were to
set the BottomLimit to 5, then any attempt to move the render object into or
beyond the 5 units at the bottom of its container would fail.</p>

<p>Limits only apply to movement, as induced through the Move() action.  This
means that limits can be over-ridden by setting the coordinate fields
directly (which can be useful in certain cases).</p>

###SEE ALSO###
Fields: LeftLimit, RightLimit, TopLimit

###END###

*****************************************************************************/

static ERROR SET_BottomLimit(struct Render *Self, LONG *Value)
{
   Self->BottomLimit = *Value;
   UpdateRenderList(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   BottomMargin
Short:  Manipulates the bottom margin of a render object.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The Render class supports margin settings, which are similar to the concept
of margins on printed paper.  Margin values have no significant meaning or
affect on a render object itself, but they are often used by other objects and
can be helpful in interface construction.  For instance, the Window template
uses margins to indicate the space available for placing graphics and other
render objects inside of it.</p>

<p>By default, all margins are set to zero when a new render object is
created.</p>

###SEE ALSO###
Fields: LeftMargin, RightMargin, TopMargin

###END###

*****************************************************************************/

static ERROR SET_BottomMargin(struct Render *Self, LONG *Value)
{
   if (*Value < 0) Self->BottomMargin = -(*Value);
   else Self->BottomMargin = *Value;
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a render object.
Type:   LONG/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The height of a render object is manipulated through this field,
although you can also use the Resize() action, which is faster if you need
to set both the Width and the Height.  You can set the Height as a fixed
value by default, or as a relative value if you set the FD_PERCENT field.
Relative heights are always calculated in relationship to a render object's
container, e.g. if the container is 200 pixels high and render Height is
80%, then your render object will be 160 pixels high.</p>

<p>Setting the Height while a render object is on display causes an
immediate graphical update to reflect the change.  Any objects that are
within the rendered area will be re-drawn and resized as necessary.</p>

<P>Height values of 0 or less are illegal, and will result in an
ERR_OutOfRange error-code.</p>

###SEE ALSO###
Fields: Width

###END###

****************************************************************************/

static ERROR GET_Height(struct Render *Self, struct Variable *Value)
{
   if (Value->Type & FD_PERCENTAGE) {
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->prvHeightPercent;
         Value->Large  = (LARGE)Self->prvHeightPercent;
         Value->Pointer = NULL;
         return(ERR_Okay);
      }
      else return(ERR_Failed);
   }
   else {
      Value->Double  = Self->Height;
      Value->Large   = Self->Height;
      Value->Pointer = NULL;
      return(ERR_Okay);
   }
}

static ERROR SET_Height(struct Render *Self, struct Variable *Variable)
{
   struct Render *parent;
   struct Variable var;
   FLOAT value;

   if (Variable->Type & FD_DOUBLE)      value = Variable->Double;
   else if (Variable->Type & FD_LARGE)  value = Variable->Large;
   else if (Variable->Type & FD_STRING) value = StrToInt(Variable->Pointer);
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (value <= 0) {
      if (Self->Head.ObjectFlags & NF_INITIALISED) return(ERR_InvalidDimension);
      else {
         Self->Dimensions &= ~(DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT);
         return(ERR_Okay);
      }
   }
   if (value > 0x7fff) value = 0x7fff;

   if (Variable->Type & FD_PERCENTAGE) {
      if (Self->ParentID) {
         if (AccessObject(Self->ParentID, 500, &parent) IS ERR_Okay) {
            Self->prvHeightPercent = value;
            Self->Dimensions    = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
            ResizeLayer(Self, Self->XCoord, Self->YCoord, 0, parent->Height * value / 100, 0, 0, 0, 0, NULL);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
      else {
         Self->prvHeightPercent = value;
         Self->Dimensions    = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      }
   }
   else {
      ResizeLayer(Self, Self->XCoord, Self->YCoord, 0, value, 0, 0, 0, 0, NULL);
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;

      /*** If the offset flags are used, adjust the vertical position ***/
      if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         var.Type   = FD_DOUBLE|FD_PERCENTAGE;
         var.Double = Self->prvYOffsetPercent;
         SET_YOffset(Self, &var);
      }
      else if (Self->Dimensions & DMF_FIXED_YOFFSET) {
         var.Type   = FD_DOUBLE;
         var.Double = Self->prvYOffset;
         SET_YOffset(Self, &var);
      }

      Action(AC_Draw, Self, NULL);
   }

   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   InsideHeight
Short:  Reflects the amount of space between the vertical margins.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>You can determine the internal height of a render object by reading the
InsideHeight field.  The returned value is the result of calculating
this formula:

<pre>
  Height - TopMargin - BottomMargin
</pre>

<p>If you have not set the TopMargin and/or BottomMargin fields, then the
returned value will be equal to the render object's height.</p>

###SEE ALSO###
Fields: InsideWidth

###END###

****************************************************************************/

static ERROR GET_InsideHeight(struct Render *Self, LONG *Value)
{
   *Value = Self->Height - Self->TopMargin - Self->BottomMargin;
   return(ERR_Okay);
}

static ERROR SET_InsideHeight(struct Render *Self, LONG *Value)
{
   LONG height;
   height = *Value + Self->TopMargin + Self->BottomMargin;
   SetField(Self, FID_Height, FT_LONG, &height);
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   InsideWidth
Short:  Reflects the amount of space between the horizontal margins.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>You can determine the internal width of a render object by reading the
InsideWidth field.  The returned value is the result of calculating
this formula:

<pre>
  Width - LeftMargin - RightMargin
</pre>

<p>If you have not set the LeftMargin and/or RightMargin fields, then the
returned value will be equal to the render object's width.</p>

###SEE ALSO###
Fields: InsideHeight

###END###

****************************************************************************/

static ERROR GET_InsideWidth(struct Render *Self, LONG *Value)
{
   *Value = Self->Width - Self->LeftMargin - Self->RightMargin;
   return(ERR_Okay);
}

static ERROR SET_InsideWidth(struct Render *Self, LONG *Value)
{
   LONG width;
   width = *Value + Self->LeftMargin + Self->RightMargin;
   SetField(Self, FID_Width, FT_LONG, &width);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   LeftLimit
Short:  Prevents a render object from moving beyond a given point on the left-hand side.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>You can prevent a render object from moving beyond a given point at the
left-hand side of its container by setting this field.  If for example you
were to set the LeftLimit to 3, then any attempt to move the render object
into or beyond the 3 units at the left of its container would fail.</p>

<p>Limits only apply to movement, as induced through the Move() action.  This
means that you can override limits by setting the coordinate fields directly
(which can be useful in certain cases).</p>

###SEE ALSO###
Fields: BottomLimit, RightLimit, TopLimit

###END###

*****************************************************************************/

static ERROR SET_LeftLimit(struct Render *Self, LONG *Value)
{
   Self->LeftLimit = *Value;
   UpdateRenderList(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   LeftMargin
Short:  Manipulates the left margin of a render object.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The Render class supports margin settings, which are similar to the concept
of margins on printed paper.  Margin values have no significant meaning or
affect on a render object itself, but they are often used by other objects and
can be helpful in interface construction.  For instance, the Window template
uses margins to indicate the space available for placing graphics and other
render objects inside of it.</p>

<p>By default, all margins are set to zero when a new render object is
created.</p>

###SEE ALSO###
Fields: BottomMargin, RightMargin, TopMargin

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   MaxHeight
Short:  Prevents the height of a render object from exceeding a certain value.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>You can limit the maximum height of a render object by setting this field.
Limiting the height affects resizing, making it impossible to use the Resize()
action to extend beyond the height you specify.</p>

<p>It is possible to circumvent the MaxHeight by setting the Height field
directly.  Note that the MaxHeight value refers to the inside-height of the
drawable area, thus the overall maximum height will include both the
TopMargin and BottomMargin values.</p>

###SEE ALSO###
Fields: Height, MaxWidth, MinHeight

###END###

*****************************************************************************/

static ERROR SET_MaxHeight(struct Render *Self, LONG *Value)
{
   Self->MaxHeight = *Value;

   if ((!Self->ParentID) AND (Self->ScreenID)) {
      struct mtScreenSizeHints hints;
      hints.MinWidth  = -1;
      hints.MinHeight = -1;
      hints.MaxWidth  = Self->MaxWidth + Self->LeftMargin + Self->RightMargin;
      hints.MaxHeight = Self->MaxHeight + Self->TopMargin + Self->BottomMargin;
      ActionMsg(MT_ScreenSizeHints, Self->ScreenID, &hints);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MaxWidth
Short:  Prevents the width of a render object from exceeding a certain value.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>You can limit the maximum width of a render object by setting this field.
Limiting the width affects resizing, making it impossible to use the Resize()
action to extend beyond the width you specify.</p>

<p>It is possible to circumvent the MaxWidth by setting the Width field
directly.  Note that the MaxWidth value refers to the inside-width of the
drawable area, thus the overall maximum width will include both the
LeftMargin and RightMargin values.</p>

###SEE ALSO###
Fields: MaxHeight, MinWidth, Width

###END###

*****************************************************************************/

static ERROR SET_MaxWidth(struct Render *Self, LONG *Value)
{
   Self->MaxWidth = *Value;

   if ((!Self->ParentID) AND (Self->ScreenID)) {
      struct mtScreenSizeHints hints;
      hints.MinWidth  = -1;
      hints.MinHeight = -1;
      hints.MaxWidth  = Self->MaxWidth + Self->LeftMargin + Self->RightMargin;
      hints.MaxHeight = Self->MaxHeight + Self->TopMargin + Self->BottomMargin;
      ActionMsg(MT_ScreenSizeHints, Self->ScreenID, &hints);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MinHeight
Short:  Prevents the height of a render object from shrinking beyond a certain value.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>You can prevent the height of a render object from shrinking too far by
setting this field.  This feature specifically affects resizing, making it
impossible to use the Resize() action to shrink the height of a render object
to a value less than the one you specify.</p>

<p>It is possible to circumvent the MinHeight by setting the Height field
directly.</p>

###SEE ALSO###
Field: MinWidth

###END###

*****************************************************************************/

static ERROR SET_MinHeight(struct Render *Self, LONG *Value)
{
   Self->MinHeight = *Value;
   if (Self->MinHeight < 1) Self->MinHeight = 1;

   if ((!Self->ParentID) AND (Self->ScreenID)) {
      struct mtScreenSizeHints hints;
      hints.MinWidth  = Self->MinWidth + Self->LeftMargin + Self->RightMargin;
      hints.MinHeight = Self->MinHeight + Self->TopMargin + Self->BottomMargin;
      hints.MaxWidth  = -1;
      hints.MaxHeight = -1;
      ActionMsg(MT_ScreenSizeHints, Self->ScreenID, &hints);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MinWidth
Short:  Prevents the width of a render object from shrinking beyond a certain value.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>You can prevent the width of a render object from shrinking too far by
setting this field.  This feature specifically affects resizing, making it
impossible to use the Resize() action to shrink the width of a render object
to a value less than the one you specify.</p>

<p>It is possible to circumvent the MinWidth by setting the Width field
directly.</p>

###SEE ALSO###
Field: MinHeight

###END###

*****************************************************************************/

static ERROR SET_MinWidth(struct Render *Self, LONG *Value)
{
   Self->MinWidth = *Value;
   if (Self->MinWidth < 1) Self->MinWidth = 1;

   if ((!Self->ParentID) AND (Self->ScreenID)) {
      struct mtScreenSizeHints hints;
      hints.MinWidth  = Self->MinWidth + Self->LeftMargin + Self->RightMargin;
      hints.MinHeight = Self->MinHeight + Self->TopMargin + Self->BottomMargin;
      hints.MaxWidth  = -1;
      hints.MaxHeight = -1;
      ActionMsg(MT_ScreenSizeHints, Self->ScreenID, &hints);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Right
Short:  Returns the right-most coordinate of a render object.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>You can read this field to find out what the right-most coordinate of
a render object is.  This is basically the opposite of the XCoord field,
and is calculated by adding the XCoord and Width fields together.</p>

###SEE ALSO###
Fields: Bottom, XCoord

###END###

*****************************************************************************/

static ERROR GET_Right(struct Render *Self, LONG *Value)
{
   *Value = Self->XCoord + Self->Width;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   RightLimit
Short:  Prevents a render object from moving beyond a given point on the right-hand side.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>You can prevent a render object from moving beyond a given point at the
right-hand side of its container by setting this field.  If for example you
were to set the RightLimit to 8, then any attempt to move the render object
into or beyond the 8 units at the right-hand side of its container would
fail.</p>

<p>Limits only apply to movement, as induced through the Move() action.  This
means that limits can be over-ridden by setting the coordinate fields
directly (which can be useful in certain cases).</p>

###SEE ALSO###
Fields: BottomLimit, LeftLimit, TopLimit

###END###

*****************************************************************************/

static ERROR SET_RightLimit(struct Render *Self, LONG *Value)
{
   Self->RightLimit = *Value;
   UpdateRenderList(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   RightMargin
Short:  Manipulates the right margin of a render object.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The Render class supports margin settings, which are similar to the concept
of margins on printed paper.  Margin values have no significant meaning or
affect on a render object itself, but they are often used by other objects and
can be helpful in interface construction.  For instance, the Window template
uses margins to indicate the space available for placing graphics and other
render objects inside of it.</p>

<p>By default, all margins are set to zero when a new render object is
created.</p>

###END###

*****************************************************************************/

static ERROR SET_RightMargin(struct Render *Self, LONG *Value)
{
   if (*Value < 0) Self->RightMargin = -*Value;
   else Self->RightMargin = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TopLimit
Short:  Prevents a render object from moving beyond a given point at the top of its container.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>You can prevent a render object from moving beyond a given point at the
top of its container by setting this field.  If for example you were to set
the TopLimit to 10, then any attempt to move the render object into or beyond
the 10 units at the top of its container would fail.</p>

<p>Limits only apply to movement, as induced through the Move() action.  This
means that limits can be over-ridden by setting the coordinate fields
directly (which can be useful in certain cases).</p>

###SEE ALSO###
Fields: BottomLimit, LeftLimit, RightLimit

###END###

*****************************************************************************/

static ERROR SET_TopLimit(struct Render *Self, LONG *Value)
{
   Self->TopLimit = *Value;
   UpdateRenderList(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TopMargin
Short:  Manipulates the top margin of a render object.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The Render class supports margin settings, which are similar to the concept
of margins on printed paper.  Margin values have no significant meaning or
affect on a render object itself, but they are often used by other objects and
can be helpful in interface construction.  For instance, the Window template
uses margins to indicate the space available for placing graphics and other
render objects inside of it.</p>

<p>By default, all margins are set to zero when a new render object is
created.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   VisibleHeight
Short:  Reflects the visible height of a rendered area within its container.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>Sometimes it can be useful to know how much of a rendered area is visible
to the user and how much is obscured or clipped by its container.  The
easiest way to ascertain the visibility level is to read the VisibleWidth and
VisibleHeight fields which do the necessary calculations for you.</p>

<p>If none of the rendered area is visible then zero is returned, otherwise
the amount of visible units are returned back to you.  The returned value
is always positive.</p>

###SEE ALSO###
Field: VisibleWidth

###END###

*****************************************************************************/

static ERROR GET_VisibleHeight(struct Render *Self, LONG *Value)
{
   struct Render *parent;

   if (!Self->ParentID) {
      *Value = Self->Height;
      return(ERR_Okay);
   }
   else if (AccessObject(Self->ParentID, 3000, &parent) IS ERR_Okay) {
      if (Self->YCoord + Self->Height > parent->Height) {
         *Value = -Self->YCoord + Self->Height;
      }
      else *Value = Self->Height;

      if (Self->YCoord < 0) *Value = *Value + Self->YCoord;
      if (*Value < 0) *Value = 0.0;

      ReleaseObject(parent);
      return(ERR_Okay);
   }
   else {
      *Value = 0.0;
      DPrintF("Get.VisibleHeight:","Failed to access parent #%d.", Self->ParentID);
      return(ERR_AccessObject);
   }
}

/*****************************************************************************

###FIELD###
Name:   VisibleWidth
Short:  Reflects the visible width of a rendered area within its container.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>Sometimes it can be useful to know how much of a rendered area is visible
to the user and how much is obscured or clipped by its container.  The
easiest way to ascertain the visibility level is to read the VisibleWidth and
VisibleHeight fields, which do the necessary calculations for you.</p>

<p>If none of the rendered area is visible then zero is returned, otherwise
the amount of visible units are returned back to you.  The returned value
is always positive.</p>

###SEE ALSO###
Field: VisibleHeight

###END###

*****************************************************************************/

static ERROR GET_VisibleWidth(struct Render *Self, LONG *Value)
{
   struct Render *parent;

   if (!Self->ParentID) {
      *Value = Self->Width;
      return(ERR_Okay);
   }
   else if (AccessObject(Self->ParentID, 2000, &parent) IS ERR_Okay) {
      if (Self->XCoord + Self->Width > parent->Width) {
         *Value = -Self->XCoord + Self->Width;
      }
      else *Value = Self->Width;

      if (Self->XCoord < 0) *Value += Self->XCoord;
      if (*Value < 0) *Value = 0.0;

      ReleaseObject(parent);
      return(ERR_Okay);
   }
   else {
      *Value = 0.0;
      DPrintF("Get.VisibleWidth:","Failed to access parent #%d.", Self->ParentID);
      return(ERR_AccessObject);
   }
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a render object.
Type:   LONG/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The width of a render object is manipulated through this field,
although you can also use the Resize() action, which is faster if you need
to set both the Width and the Height.  You can set the Width as a fixed
value by default, or as a relative value if you set the FD_PERCENT field.
Relative widths are always calculated in relationship to a render object's
container, e.g. if the container is 200 pixels wide and render Width is
80%, then your render object will be 160 pixels wide.</p>

<p>Setting the Width while a render object is on display causes an
immediate graphical update to reflect the change.  Any objects that are
within the rendered area will be re-drawn and resized as necessary.</p>

<P>Width values of 0 or less are illegal, and will result in an
ERR_OutOfRange error-code.</p>

###SEE ALSO###
Field; Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct Render *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Type & FD_PERCENTAGE) {
         if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
            Value->Double = Self->prvWidthPercent;
         }
         else return(ERR_Failed);
      }
      else Value->Double = Self->Width;
   }
   else if (Value->Type & FD_LARGE) {
      Value->Large = (LARGE)Self->Width;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Width(struct Render *Self, struct Variable *Variable)
{
   struct Variable var;
   struct Render *parent;
   FLOAT value;

   if (Variable->Type & FD_DOUBLE)      value = Variable->Double;
   else if (Variable->Type & FD_LARGE)  value = Variable->Large;
   else if (Variable->Type & FD_STRING) value = StrToInt(Variable->Pointer);
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (value <= 0) {
      if (Self->Head.ObjectFlags & NF_INITIALISED) return(ERR_InvalidDimension);
      else {
         Self->Dimensions &= ~(DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH);
         return(ERR_Okay);
      }
   }
   if (value > 0x7fff) value = 0x7fff;

   if (Variable->Type & FD_PERCENTAGE) {
      if (Self->ParentID) {
         if (AccessObject(Self->ParentID, 500, &parent) IS ERR_Okay) {
            Self->prvWidthPercent = value;
            Self->Dimensions   = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
            ResizeLayer(Self, Self->XCoord, Self->YCoord, parent->Width * value / 100, 0, 0, 0, 0, 0, NULL);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
      else {
         Self->prvWidthPercent = value;
         Self->Dimensions   = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      }
   }
   else {
      ResizeLayer(Self, Self->XCoord, Self->YCoord, value, 0, 0, 0, 0, 0, NULL);
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;

      /*** If the offset flags are used, adjust the horizontal position ***/
      if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         var.Type = FD_DOUBLE|FD_PERCENTAGE;
         var.Double = Self->prvXOffsetPercent;
         SET_XOffset(Self, &var);
      }
      else if (Self->Dimensions & DMF_FIXED_XOFFSET) {
         var.Type = FD_DOUBLE;
         var.Double = Self->prvXOffset;
         SET_XOffset(Self, &var);
      }

      Action(AC_Draw, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name;     XCoord
Synonyms: X
Short:    Determines the horizontal position of a render object.
Type:     LONG/PERCENTAGE
Status:   Read/Write

###DESCRIPTION###
<p>The horizontal position of a render object can be set through this field.
You have the choice of setting a fixed coordinate (the default) or a
relative coordinate if you use the FD_PERCENT flag.</p>

<p>If you set the XCoord while the render object is on display, the position
of the rendered area will be updated immediately.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_XCoord(struct Render *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Type & FD_PERCENTAGE) Value->Double = Self->prvXPercent;
      else Value->Double = Self->XCoord;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Type & FD_PERCENTAGE) Value->Large = (LARGE)Self->prvXPercent;
      else Value->Large = Self->XCoord;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XCoord(struct Render *Self, struct Variable *Variable)
{
   struct Render *Container;
   FLOAT value;

   if (Variable->Type & FD_DOUBLE)      value = Variable->Double;
   else if (Variable->Type & FD_LARGE)  value = Variable->Large;
   else if (Variable->Type & FD_STRING) value = StrToInt(Variable->Pointer);
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Variable->Type & FD_PERCENTAGE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
      Self->prvXPercent   = value;
      if (Self->ParentID) {
         if (AccessObject(Self->ParentID, 500, &Container) IS ERR_Okay) {
            MoveLayer(Self, (Container->Width * value) / 100, Self->YCoord);
            ReleaseObject(Container);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
   }
   else {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
      MoveLayer(Self, value, Self->YCoord);

      /* If our right-hand side is relative, we need to resize our
      ** drawable to counteract the movement.
      */

      if ((Self->ParentID) AND (Self->Dimensions & (DMF_RELATIVE_XOFFSET|DMF_FIXED_XOFFSET))) {
         if (AccessObject(Self->ParentID, 1000, &Container) IS ERR_Okay) {
            ResizeLayer(Self, Self->XCoord, Self->YCoord, Container->Width - Self->XCoord - Self->prvXOffset, 0, 0, 0, 0, 0, NULL);
            ReleaseObject(Container);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Determines the horizontal offset of a render object.
Type:   LONG/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with the XCoord or Width fields.</p>

<p>If set in conjunction with the XCoord field, the width of the render object
will be from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the render object
is dynamically calculated in relation to the width of its container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the render object will be positioned at an X coordinate calculated from
the formula "XCoord = ContainerWidth - RenderWidth - XOffset".</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Render *Self, struct Variable *Variable)
{
   struct Variable xoffset;
   struct Render *parent;
   DOUBLE value;

   if (Variable->Type & FD_PERCENTAGE) {
      xoffset.Type = FD_DOUBLE;
      if (GET_XOffset(Self, &xoffset) IS ERR_Okay) {
         value = xoffset.Double / Self->Width * 100;
      }
      else value = 0;
   }
   else {
      if (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) {
         value = Self->prvXOffset;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->ParentID)) {
         if (AccessObject(Self->ParentID, 1000, &parent) IS ERR_Okay) {
            value = parent->Width - Self->XCoord - Self->Width;
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_GetDimension, ERR_ExclusiveDenied));
      }
      else value = 0;
   }

   if (Variable->Type & FD_DOUBLE)     Variable->Double = value;
   else if (Variable->Type & FD_LARGE) Variable->Large  = value;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Render *Self, struct Variable *Variable)
{
   struct Render *parent;
   FLOAT value;

   if (Variable->Type & FD_DOUBLE)      value = Variable->Double;
   else if (Variable->Type & FD_LARGE)  value = Variable->Large;
   else if (Variable->Type & FD_STRING) value = StrToInt(Variable->Pointer);
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (value < 0) value = -value;

   if (Variable->Type & FD_PERCENTAGE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
      Self->prvXOffsetPercent = value;

      if (Self->ParentID) {
         if (AccessObject(Self->ParentID, 500, &parent) IS ERR_Okay) {
            Self->prvXOffset = (parent->Width * Self->prvXOffsetPercent) / 100;
            if (!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X))) Self->XCoord = parent->Width - Self->prvXOffset - Self->Width;
            if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
               ResizeLayer(Self, Self->XCoord, Self->YCoord, parent->Width - Self->XCoord - Self->prvXOffset, 0, 0, 0, 0, 0, NULL);
            }
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
   }
   else {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
      Self->prvXOffset = value;

      if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND (Self->ParentID)) {
         if (AccessObject(Self->ParentID, 1000, &parent) IS ERR_Okay) {
            MoveLayer(Self, parent->Width - Self->prvXOffset - Self->Width, Self->YCoord);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND (Self->ParentID)) {
         if (AccessObject(Self->ParentID, 1000, &parent) IS ERR_Okay) {
            ResizeLayer(Self, Self->XCoord, Self->YCoord, parent->Width - Self->XCoord - Self->prvXOffset, 0, 0, 0, 0, 0, NULL);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:     YCoord
Synonyms: Y
Short:    Determines the vertical position of a render object.
Type:     LONG/PERCENTAGE
Status:   Read/Write

###DESCRIPTION###
<p>The vertical position of a render object can be set through this field.
You have the choice of setting a fixed coordinate (the default) or a relative
coordinate if you use the FD_PERCENT flag.</p>

<p>If you set the YCoord while the render object is on display, the position
of the rendered area will be updated immediately.</p>

###SEE ALSO###
Fields: XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_YCoord(struct Render *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Type & FD_PERCENTAGE) Value->Double = Self->prvYPercent;
      else Value->Double = Self->YCoord;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Type & FD_PERCENTAGE) Value->Large = (LARGE)Self->prvYPercent;
      else Value->Large = Self->YCoord;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YCoord(struct Render *Self, struct Variable *Variable)
{
   struct Render *parent;
   FLOAT value;

   if (Variable->Type & FD_DOUBLE)     value = Variable->Double;
   else if (Variable->Type & FD_LARGE) value = Variable->Large;
   else if (Variable->Type & FD_STRING) value = StrToInt(Variable->Pointer);
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Variable->Type & FD_PERCENTAGE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
      Self->prvYPercent   = value;
      if (Self->ParentID) {
         if (AccessObject(Self->ParentID, 500, &parent) IS ERR_Okay) {
            MoveLayer(Self, Self->XCoord, (parent->Height * value) / 100);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
   }
   else {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
      MoveLayer(Self, Self->XCoord, value);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Determines the vertical offset of a render object.
Type:   LONG/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with the YCoord or Height fields.</p>

<p>If set in conjunction with the YCoord field, the height of the render object
will be from that Y coordinate up to the height of the container, minus the
value given in the YOffset.  This means that the height of the render object
is dynamically calculated in relation to the height of its container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the render object will be positioned at a Y coordinate calculated from
the formula "YCoord = ContainerHeight - RenderHeight - YOffset".</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Render *Self, struct Variable *Variable)
{
   struct Variable yoffset;
   struct Render *parent;
   DOUBLE value;

   if (Variable->Type & FD_PERCENTAGE) {
      yoffset.Type = FD_DOUBLE;
      if (GET_YOffset(Self, &yoffset) IS ERR_Okay) {
         value = yoffset.Double / Self->Height * 100;
      }
      else value = 0;
   }
   else {
      if (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) {
         value = Self->prvYOffset;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->ParentID)) {
         if (AccessObject(Self->ParentID, 1000, &parent) IS ERR_Okay) {
            value = parent->Height - Self->YCoord - Self->Height;
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
      else value = 0;
   }

   if (Variable->Type & FD_DOUBLE)     Variable->Double = value;
   else if (Variable->Type & FD_LARGE) Variable->Large  = value;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Render *Self, struct Variable *Variable)
{
   struct Render *parent;
   FLOAT value;

   if (Variable->Type & FD_DOUBLE)      value = Variable->Double;
   else if (Variable->Type & FD_LARGE)  value = Variable->Large;
   else if (Variable->Type & FD_STRING) value = StrToInt(Variable->Pointer);
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (value < 0) value = -value;

   if (Variable->Type & FD_PERCENTAGE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
      Self->prvYOffsetPercent = value;

      if (Self->ParentID) {
         if (AccessObject(Self->ParentID, 500, &parent) IS ERR_Okay) {
            Self->prvYOffset = (parent->Height * Self->prvYOffsetPercent) / 100;
            if (!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)))Self->YCoord = parent->Height - Self->prvYOffset - Self->Height;
            if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
               ResizeLayer(Self, Self->XCoord, Self->YCoord, 0, parent->Height - Self->YCoord - Self->prvYOffset, 0, 0, 0, 0, NULL);
            }
            else MoveLayer(Self, Self->XCoord, parent->Height - Self->prvYOffset - Self->Height);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
   }
   else {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
      Self->prvYOffset = value;

      if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND (Self->ParentID)) {
         if (AccessObject(Self->ParentID, 1000, &parent) IS ERR_Okay) {
            if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
               ResizeLayer(Self, Self->XCoord, Self->YCoord, 0, parent->Height - Self->YCoord - Self->prvYOffset, 0, 0, 0, 0, NULL);
            }
            else MoveLayer(Self, Self->XCoord, parent->Height - Self->prvYOffset - Self->Height);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND (Self->ParentID)) {
         if (AccessObject(Self->ParentID, 1000, &parent) IS ERR_Okay) {
            ResizeLayer(Self, Self->XCoord, Self->YCoord, 0, parent->Height - Self->YCoord - Self->prvYOffset, 0, 0, 0, 0, NULL);
            ReleaseObject(parent);
         }
         else return(ObjectError(Self, ERH_SetDimension, ERR_ExclusiveDenied));
      }
   }
   return(ERR_Okay);
}

