/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Scroll
Version:   1.0
Status:    Stable
ID:        ID_SCROLL
Category:  GUI
Date:      December 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1999-2003.  All rights reserved.
Keywords:  Scroll, ScrollBar, Scrolling, Slider
Short:     This class is used to aid in the creation of scrollbars and sliders.

###DESCRIPTION###
<p>The Scroll class provides a facility for creating scrollbars that allow
the user to move drawable objects within their containers.  It can also be
used to scroll contents or 'pages' of information (useful for Text Editors
for instance).  In either case, it is most useful in situations  where the
available graphics space is not sufficient for the amount of information
that needs to be shown.  The Scroll class in itself is only designed to
provide scrolling functionality, i.e. it does not create any gadgetry or
graphics for the scrollbar.  For that reason, complete scrollbars are usually
created using scripts or helper classes, such as the ScrollBar class.</p>

<p>The Scroll class is designed to provide scrolling in 3 different
directions - along the X, Y, or Z axis.  You can combine the different
axis, so a diagonal scrolling gadget could be created for example.  It
is also possible to create buttons that are attached to the scroll object.
Refer to the Left, Right, Up, Down, In and Out fields for more
information.</p>

<p>You can use the Scroll class to create fixed or proportional scrollbars
depending on what the situation dictates.  Where possible you should use
proportional scrolling as it is the easier of the two to set up.  All you
need to do is specify the page size (which represents the width or height of
the graphical content) and the view size (the 'window' into the information).
If you are going to create a fixed scrollbar, you will need to calculate the
correct scale based on the size of the scrollbar and the area that is to
be scrolled.  Refer to the Scale field for more information on how this
works.</p>

<p>Some objects are supportive of scrolling - for example, the Text
class supports horizontal and vertical scrolling through its HScroll and
VScroll fields.  Objects that support scrollbars expect to be connected
directly to an appropriate scroll object.  Once that connection is
established, they will take over the scroll object so that the page size
and positioning is always managed correctly.  In such cases it is not
necessary for you to pre-calculate the scale or page and view sizes.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/scroll.h>
#include <pandora/modules/render.h>

//#define DEBUG TRUE

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_SCROLL  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_SCROLL, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1999-2004.  All rights reserved.",
   "August 2004",
   "Scroll"
};

static struct KernelBase *KernelBase;
static struct RenderBase *RenderBase;
static OBJECTPTR modRender = NULL;
static OBJECTPTR ScrollClass = NULL;

/*****************************************************************************
** Scroll class structure.
*/

struct ScrollButton {
   OBJECTID ButtonID;
   BYTE Direction;
};

struct Scroll {
   OBJECT_HEADER           /* Class header */
   OBJECTID ObjectID;      /* Object to receive the scroll action */
   OBJECTID SliderID;      /* Drawable that represents the slider within the scrollbar gadget */
   OBJECTID IntersectID;   /* Intersecting scrollbar, if relevant */
   OBJECTID MonitorID;     /* Drawable to monitor for extra scroll commands (e.g. from the mouse wheel) */
   LONG     Unit;          /* Amount of units to move when buttons are pressed */
   LONG     Flags;         /* Special flags */
   LONG     PageSize;      /* Size of the page (typically larger than the view) */
   LONG     ViewSize;      /* Size of the view (contains the page) */
   LONG     Position;      /* Page offset */
   LONG     StartMargin;
   LONG     EndMargin;
   LONG     SliderSize;
   LONG     Axis;

   /*** Private fields ***/
   OBJECTID ScrollbarID;       /* Drawable that represents the whole of the scroll bar */
   OBJECTID IntersectDrawable; /* Drawable that represents the intersecting scrollbar, if there is one */
   LONG     BarSize;           /* Indicates the amount of pixel space within the scrollbar area */
   LONG     PrevCoord;
   LONG     SliderX;
   LONG     SliderY;
   LONG     ObscuredView;
   LONG     DrawableWidth;
   LONG     DrawableHeight;
   FIELD    FieldID;
   UBYTE    RecursionBlock;
   BYTE     PostIntersect;     /* Set to TRUE if the Intersect field was set post-initialisation (we are not expected to adjust our scrollbar size) */
   struct ScrollButton Buttons[8];
};

/*****************************************************************************
** Scroll flags.
*/

static struct FieldDef ScrollFlags[] = {
   { "AUTOHIDE",     SCF_AUTOHIDE },
   { "HORIZONTAL",   SCF_HORIZONTAL },
   { "VERTICAL",     SCF_VERTICAL },
   { "REVERSE",      SCF_REVERSE },
   { "RELATIVE",     SCF_RELATIVE },
   { "SLIDER",       SCF_SLIDER },
   { "FIXED",        SCF_FIXED },
   { "AUTOACTIVATE", SCF_AUTOACTIVATE },
   { NULL, NULL }
};

/*****************************************************************************
** Axis options.
*/

static struct FieldDef AxisLookup[] = {
   { "X", AXIS_X },
   { "Y", AXIS_Y },
   { "Z", AXIS_Z },
   { NULL, NULL }
};

/*****************************************************************************
** Scroll class definition.
*/

static ERROR SET_Axis(struct Scroll *, LONG *);
static ERROR SET_Field(struct Scroll *, STRING);
static ERROR SET_Intersect(struct Scroll *, OBJECTID *);
static ERROR SET_Monitor(struct Scroll *, LONG *);
static ERROR SET_PageSize(struct Scroll *, LONG *);
static ERROR SET_Position(struct Scroll *, LONG *);
static ERROR SET_SliderSize(struct Scroll *, LONG *);
static ERROR SET_ViewSize(struct Scroll *, LONG *);

static struct FieldArray ScrollFields[] = {
   { "Object",      0, FDF_OBJECTID|FDF_RW,  0, NULL, NULL },
   { "Slider",      0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Intersect",   0, FDF_OBJECTID|FDF_RW,  ID_SCROLL, NULL, SET_Intersect },
   { "Monitor",     0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, SET_Monitor },
   { "Unit",        0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Flags",       0, FDF_LONGFLAGS|FDF_RW, (LONG)&ScrollFlags, NULL, NULL },
   { "PageSize",    0, FDF_LONG|FDF_RW,      0, NULL, SET_PageSize },
   { "ViewSize",    0, FDF_LONG|FDF_RW,      0, NULL, SET_ViewSize },
   { "Position",    0, FDF_LONG|FDF_RW,      0, NULL, SET_Position },
   { "StartMargin", 0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "EndMargin",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "SliderSize",  0, FDF_LONG|FDF_RI,      0, NULL, SET_SliderSize },
   { "Axis",        0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&AxisLookup, NULL, SET_Axis },
   /*** Virtual Fields ***/
   { "Field",       0, FDF_STRING|FDF_W,     0, NULL, SET_Field },
   END_FIELD
};

/*****************************************************************************
** Scroll actions.
*/

static ERROR SCROLL_ActionNotify(struct Scroll *, struct acActionNotify *);
static ERROR SCROLL_Activate(struct Scroll *, APTR);
static ERROR SCROLL_Init(struct Scroll *, APTR);
static ERROR SCROLL_NewObject(struct Scroll *, APTR);

static struct ActionArray ScrollActions[] = {
   { AC_ActionNotify, SCROLL_ActionNotify },
   { AC_Activate,     SCROLL_Activate },
   { AC_Init,         SCROLL_Init },
   { AC_NewObject,    SCROLL_NewObject },
   { NULL, NULL }
};

/*****************************************************************************
** Scroll methods.
*/

static ERROR SCROLL_AddButton(struct Scroll *, struct mtAddScrollButton *);
static ERROR SCROLL_UpdateScroll(struct Scroll *, struct mtUpdateScroll *);

static struct FunctionField argsUpdateScroll[] = { { "PageSize", ARG_LONG }, { "ViewSize", ARG_LONG }, { "Position", ARG_LONG }, { "Unit", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsAddButton[]    = { { "Button", ARG_LONG }, { "Direction", ARG_LONG }, { NULL, NULL } };

static struct MethodArray ScrollMethods[] = {
   { MT_UpdateScroll,    SCROLL_UpdateScroll, "UpdateScroll", argsUpdateScroll, sizeof(struct mtUpdateScroll) },
   { MT_AddScrollButton, SCROLL_AddButton,    "AddButton",    argsAddButton,    sizeof(struct mtAddScrollButton) },
   { NULL, NULL, NULL, NULL, NULL }
};

/****************************************************************************/

static ERROR ProcessClick(struct Scroll *Self, OBJECTID NotifyID, LONG XCoord, LONG YCoord);

/****************************************************************************/

static inline LONG check_position(struct Scroll *Self, LONG Position)
{
   if (Position < 0) return(0);
   else if (Self->PageSize <= Self->ViewSize) Position = 0;
   else if (Self->Flags & SCF_SLIDER) {
      if (Position > Self->PageSize) return(Self->PageSize);
   }
   else if (Position > Self->PageSize - Self->ViewSize) return(Self->PageSize - Self->ViewSize);

   return(Position);
}

/****************************************************************************/

static inline void set_position(struct Scroll *Self, LONG Position)
{
#ifdef DEBUG
   DPrintF("Scroll:","set_postition(%d)", Position);
#endif

   Self->Position = Position;
   if (Self->Flags & SCF_AUTOACTIVATE) acActivate(Self);

   /*** Inform the object if it wants a field update ***/

   if ((Self->FieldID) AND (Self->ObjectID)) {
      OBJECTPTR object;

      if (AccessObject(Self->ObjectID, 5000, &object) IS ERR_Okay) {
         SetField(object, Self->FieldID, FT_LONG, &Position);
         ReleaseObject(object);
      }
   }
}

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ScrollClass, NULL,
      FID_BaseClassID|TLONG, ID_SCROLL,
      FID_Name|TSTRING,      "Scroll",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      ScrollActions,
      FID_Methods|TPTR,      ScrollMethods,
      FID_Fields|TPTR,       ScrollFields,
      FID_Size|TLONG,        sizeof(struct Scroll),
      FID_Version|TFLOAT,    VER_SCROLL,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (ScrollClass) { acFree(ScrollClass); ScrollClass = NULL; }
   if (modRender)   { acFree(modRender);   modRender = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: UpdateBarSize()
** Short:    Updates the BarSize field with the current available scroll space.
*/

ERROR UpdateBarSize(struct Scroll *Self)
{
   OBJECTPTR bar;

   if (!Self->ScrollbarID) return(ERR_FieldNotSet);

   if (AccessObject(Self->ScrollbarID, 2000, &bar) IS ERR_Okay) {
      if (Self->Flags & SCF_VERTICAL) GetField(bar, FID_Height, FT_LONG, &Self->BarSize);
      else if (Self->Flags & SCF_HORIZONTAL) GetField(bar, FID_Width, FT_LONG, &Self->BarSize);

      Self->BarSize -= (Self->StartMargin + Self->EndMargin);
      ReleaseObject(bar);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************
** Action: ActionNotify()
*/

static ERROR SCROLL_ActionNotify(struct Scroll *Self, struct acActionNotify *Notify)
{
   if (Notify->Error != ERR_Okay) return(ERR_Okay);

   if (Notify->ActionID IS AC_Redimension) {
      struct mtUpdateScroll scroll;
      struct acRedimension *redimension;

      if (Self->PageSize <= 0) return(ERR_Okay);

      redimension = (struct acRedimension *)Notify->Args;

      scroll.Position = Self->Position;
      scroll.PageSize = Self->PageSize;

      if (Self->Flags & SCF_SLIDER) {
         /*** For sliders, the viewsize is preset to a fixed value ***/
         if (Self->Flags & SCF_VERTICAL) Self->BarSize = redimension->Height - (Self->StartMargin + Self->EndMargin);
         else if (Self->Flags & SCF_HORIZONTAL) Self->BarSize = redimension->Width - (Self->StartMargin + Self->EndMargin);
         scroll.ViewSize = Self->ViewSize;
      }
      else if (Self->Flags & SCF_VERTICAL) {
         Self->BarSize   = redimension->Height - (Self->StartMargin + Self->EndMargin);
         scroll.ViewSize = redimension->Height - Self->ObscuredView;
      }
      else if (Self->Flags & SCF_HORIZONTAL) {
         Self->BarSize   = redimension->Width - (Self->StartMargin + Self->EndMargin);
         scroll.ViewSize = redimension->Width - Self->ObscuredView;
      }

      if ((scroll.PageSize <= scroll.ViewSize) AND (Self->Position > 0) AND (!(Self->Flags & SCF_SLIDER))) {
         /* If the page is smaller than the view area, reset the object
         ** to position zero.
         */

         if (Self->Flags & SCF_RELATIVE) {

         }
         else {
            struct acScrollToPoint scrollto;

            scrollto.XCoord = 0;
            scrollto.YCoord = 0;
            scrollto.ZCoord = 0;
            if (Self->Axis IS AXIS_X) scrollto.Flags = STF_X;
            else scrollto.Flags = STF_Y;

            DelayMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto); /* Use a delay to give good redraw results */
         }
         scroll.Position = 0;
      }

      scroll.Unit = Self->Unit;
      SCROLL_UpdateScroll(Self, &scroll);
   }
   else if (Notify->ActionID IS AC_MoveToPoint) {
      struct acMoveToPoint *moveto;
      struct acScroll scroll;
      struct acScrollToPoint scrollto;
      LONG position, coord;

      moveto = (struct acMoveToPoint *)Notify->Args;

      /*** MoveToPoint messages originate from the slider ***/

      if (Self->Flags & SCF_HORIZONTAL) coord = moveto->XCoord;
      else coord = moveto->YCoord;

      if (Self->Flags & SCF_REVERSE) {
         if (Self->Flags & SCF_SLIDER) position = Self->PageSize * ((DOUBLE)(coord - Self->StartMargin) / (DOUBLE)(Self->BarSize - Self->SliderSize));
         else position = (Self->PageSize - Self->ViewSize) * ((DOUBLE)(coord - Self->StartMargin) / (DOUBLE)(Self->BarSize - Self->SliderSize));
         position = Self->PageSize - position;
      }
      else {
         if (Self->Flags & SCF_SLIDER) position = Self->PageSize * ((DOUBLE)(coord - Self->StartMargin) / (DOUBLE)(Self->BarSize - Self->SliderSize));
         else position = (Self->PageSize - Self->ViewSize) * ((DOUBLE)(coord - Self->StartMargin) / (DOUBLE)(Self->BarSize - Self->SliderSize));
      }

      #ifdef DEBUG
         DPrintF("ActionNotify:","[Scroll] Movement detected in slider - coord %d, position %d.", coord, position);
      #endif

      position = check_position(Self, position);

      if (position IS Self->Position) return(ERR_Okay);

      if (Self->RecursionBlock) {
         /*** Recursive block is active... ***/
         #ifdef DEBUG
            DPrintF("ActionNotify:","[Scroll] Recursive block protection.");
         #endif
         Self->Position = position;
         Self->PrevCoord = coord; /* Remember the slider position so that relative scrolling won't get out of sync */
         return(ERR_Okay);
      }

      Self->RecursionBlock++;

      if (Self->ObjectID) {
         if (Self->Flags & SCF_RELATIVE) {
            if (Self->Axis IS AXIS_X) scroll.XChange = (coord - Self->PrevCoord);
            else scroll.XChange = 0;

            if (Self->Axis IS AXIS_Y) scroll.YChange = (coord - Self->PrevCoord);
            else scroll.YChange = 0;

            if (Self->Axis IS AXIS_Z) scroll.ZChange = (coord - Self->PrevCoord);
            else scroll.ZChange = 0;

            scroll.XChange = scroll.XChange * (Self->PageSize / Self->ViewSize);
            scroll.YChange = scroll.YChange * (Self->PageSize / Self->ViewSize);
            scroll.ZChange = scroll.ZChange * (Self->PageSize / Self->ViewSize);
            ActionMsg(AC_Scroll, Self->ObjectID, &scroll);
         }
         else if (Self->Axis IS AXIS_X) {
            scrollto.XCoord = position;
            scrollto.YCoord = 0;
            scrollto.ZCoord = 0;
            scrollto.Flags  = STF_X;
            ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
         }
         else if (Self->Axis IS AXIS_Y) {
            scrollto.XCoord = 0;
            scrollto.YCoord = position;
            scrollto.ZCoord = 0;
            scrollto.Flags  = STF_Y;
            ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
         }
         else if (Self->Axis IS AXIS_Z) {
            scrollto.XCoord = 0;
            scrollto.YCoord = 0;
            scrollto.ZCoord = position;
            scrollto.Flags  = STF_Z;
            ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
         }
         else DPrintF("@Scroll:","Invalid Axis setting of %d.", Self->Axis);
      }

      set_position(Self, position);

      Self->RecursionBlock--;

      Self->PrevCoord = coord;
   }
   else if (Notify->ActionID IS AC_Hide) {
      struct mtUpdateScroll update;
      OBJECTPTR bar, intersect;
      LONG offset;

      /* The Hide action is received when an -intersecting- scrollbar is hidden.
      ** This code will adjust our position to deal with the intersection
      ** point.
      */

      if (Self->PostIntersect) {
         /*** Recompute the viewable area ***/

         Self->ObscuredView = 0;

         update.Position = Self->Position;
         update.ViewSize = Self->BarSize + Self->StartMargin + Self->EndMargin;
         update.PageSize = Self->PageSize;
         update.Unit     = Self->Unit;
         SCROLL_UpdateScroll(Self, &update);
      }
      else if (AccessObject(Self->ScrollbarID, 5000, &bar) IS ERR_Okay) {
         if (AccessObject(Self->IntersectDrawable, 5000, &intersect) IS ERR_Okay) {
            if (Self->Flags & SCF_HORIZONTAL) {
               GetFields(intersect, FID_XOffset|TLONG, &offset, TAGEND);
               SetFields(bar, FID_XOffset|TLONG, offset, TAGEND);
            }
            else if (Self->Flags & SCF_VERTICAL) {
               GetFields(intersect, FID_YOffset|TLONG, &offset, TAGEND);
               SetFields(bar, FID_YOffset|TLONG, offset, TAGEND);
            }
            ReleaseObject(intersect);
         }
         ReleaseObject(bar);
      }
   }
   else if (Notify->ActionID IS AC_Scroll) {
      struct acMove move;
      struct acScroll *scroll;

      if (Self->RecursionBlock) return(ERR_Okay);
      scroll = (struct acScroll *)Notify->Args;

      if (Notify->ObjectID IS Self->ObjectID) {
         /* If the message came from the object maintained by the scrollbar,
         ** we need to adjust our slider rather than send another scroll signal.
         */

         #ifdef DEBUG
            DPrintF("ActionNotify:","[Scroll] Scroll action received from #%d - moving the slider.", Notify->ObjectID);
         #endif

         Self->RecursionBlock++;

         move.XChange = ((DOUBLE)(scroll->XChange * Self->BarSize) / (DOUBLE)(Self->PageSize - Self->ViewSize));
         move.YChange = ((DOUBLE)(scroll->YChange * Self->BarSize) / (DOUBLE)(Self->PageSize - Self->ViewSize));
         move.ZChange = scroll->ZChange;
         ActionMsg(AC_Move, Self->SliderID, &move);

         Self->RecursionBlock--;
      }
      else if (Notify->ObjectID IS Self->MonitorID) {
         struct mtUpdateScroll update;
         struct acScrollToPoint scrollto;
         LONG position;

         #ifdef DEBUG
            DPrintF("~ActionNotify:","[Scroll] Scroll action received from monitored #%d - sending scroll signal.", Notify->ObjectID);
         #endif

         /* A scroll request has come from the monitored object.  We have to send
         ** a scroll message to the object that our scrollbar is controlling,
         ** then update our slider so that it reflects the new position.
         */

         if (Self->Axis IS AXIS_X) {
            position = check_position(Self, Self->Position + scroll->XChange);

            if (Self->ObjectID) {
               if (Self->Flags & SCF_RELATIVE) {
                  ActionMsg(AC_Scroll, Self->ObjectID, scroll); /* Pass the original scroll message straight to the subscribed object */
               }
               else {
                  scrollto.XCoord = position;
                  scrollto.YCoord = 0;
                  scrollto.ZCoord = 0;
                  scrollto.Flags  = STF_X;
                  ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
               }
            }
         }
         else if (Self->Axis IS AXIS_Y) {
            position = check_position(Self, Self->Position + scroll->YChange);

            if (Self->ObjectID) {
               scrollto.XCoord = 0;
               scrollto.YCoord = position;
               scrollto.ZCoord = 0;
               scrollto.Flags  = STF_Y;
               ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
            }
         }
         else if (Self->Axis IS AXIS_Z) {
            position = check_position(Self, Self->Position + scroll->YChange);

            if (Self->ObjectID) {
               scrollto.XCoord = 0;
               scrollto.YCoord = 0;
               scrollto.ZCoord = position;
               scrollto.Flags  = STF_Z;
               ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
            }
         }
         else {
            DPrintF("@Scroll:","Invalid Axis setting of %d.", Self->Axis);
            return(ERR_Okay);
         }

         #ifdef DEBUG
            DPrintF("ActionNotify:","[Scroll] Updating slider position.", Notify->ObjectID);
         #endif

         update.Position = position;
         update.ViewSize = Self->ViewSize;
         update.PageSize = Self->PageSize;
         update.Unit     = Self->Unit;
         SCROLL_UpdateScroll(Self, &update);

         #ifdef DEBUG
            StepBack();
         #endif
      }
   }
   else if (Notify->ActionID IS AC_Show) {
      struct mtUpdateScroll update;
      OBJECTPTR bar, intersect;
      LONG offset, size;

      /* The Show action is received when an intersecting scrollbar is shown.
      ** This code will adjust our position to deal with the intersection
      ** point.
      */

      if (Self->PostIntersect) {
         /*** Recompute the viewable area ***/
         if (AccessObject(Self->IntersectDrawable, 5000, &intersect) IS ERR_Okay) {
            LONG offset;
            if (Self->Flags & SCF_HORIZONTAL) GetFields(intersect, FID_Width|TLONG, &Self->ObscuredView, FID_XOffset|TLONG, &offset, TAGEND);
            else if (Self->Flags & SCF_VERTICAL) GetFields(intersect, FID_Height|TLONG, &Self->ObscuredView, FID_YOffset|TLONG, &offset, TAGEND);
            else { Self->ObscuredView = 0; offset = 0; }

            Self->ObscuredView += offset;

            update.Position = Self->Position;
            update.ViewSize = Self->ViewSize - Self->ObscuredView;
            update.PageSize = Self->PageSize;
            update.Unit     = Self->Unit;
            SCROLL_UpdateScroll(Self, &update);

            ReleaseObject(intersect);
         }
      }
      else if (AccessObject(Self->ScrollbarID, 5000, &bar) IS ERR_Okay) {
         if (AccessObject(Self->IntersectDrawable, 5000, &intersect) IS ERR_Okay) {
            if (Self->Flags & SCF_HORIZONTAL) {
               GetFields(intersect, FID_XOffset|TLONG, &offset, FID_Width|TLONG, &size, TAGEND);
               SetFields(bar, FID_XOffset|TLONG, offset + size, TAGEND);
            }
            else if (Self->Flags & SCF_VERTICAL) {
               GetFields(intersect, FID_YOffset|TLONG, &offset, FID_Height|TLONG, &size, TAGEND);
               SetFields(bar, FID_YOffset|TLONG, offset + size, TAGEND);
            }
            ReleaseObject(intersect);
         }
         ReleaseObject(bar);
      }
   }
   else if (Notify->ActionID IS AC_UserClickRepeat) {
      struct acUserClickRepeat *repeat;
      repeat = (struct acUserClickRepeat *)Notify->Args;
      ProcessClick(Self, Notify->ObjectID, repeat->XCoord, repeat->YCoord);
   }
   else if (Notify->ActionID IS AC_UserClick) {
      struct acUserClick *click;
      click = (struct acUserClick *)Notify->Args;
      ProcessClick(Self, Notify->ObjectID, click->XCoord, click->YCoord);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Activate()
*/

static ERROR SCROLL_Activate(struct Scroll *Self, APTR Void)
{
   LONG count;
   WORD i;

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];
      if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
         if (Self->Flags & SCF_MESSAGE) {
            for (i=0; i < count; i++) DelayMsg(AC_Activate, list[i].ObjectID, NULL);
         }
         else for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ProcessClick()
*/

#define PAGESCROLL 0.80

enum {
   DIR_NONE=0,
   DIR_NEGATIVE,
   DIR_POSITIVE
};

static ERROR ProcessClick(struct Scroll *Self, OBJECTID NotifyID, LONG XCoord, LONG YCoord)
{
   struct mtUpdateScroll update;
   struct acMove move;
   struct acScroll scroll;
   struct acScrollToPoint scrollto;
   struct RenderInfo slider;
   WORD i;
   BYTE dir;

   if (GetRenderInfo(Self->SliderID, &slider, sizeof(slider)) != ERR_Okay) return(ERR_Failed);

#ifdef DEBUG
   DPrintF("~ProcessClick()","[Scroll] %dx%d, Slider: %dx%d,%dx%d, Margins: %d,%d", XCoord, YCoord, slider.XCoord, slider.YCoord, slider.Width, slider.Height, Self->StartMargin, Self->EndMargin);
#endif

   move.XChange = 0;
   move.YChange = 0;
   move.ZChange = 0;

   if (NotifyID IS Self->ScrollbarID) {
      /* The empty area surrounding the slider was clicked.  Scroll a single
      ** page in the correct direction.
      */

      if (Self->Flags & SCF_HORIZONTAL) {
         move.XChange = slider.Width;

         if ((YCoord >= slider.YCoord) AND (YCoord <= slider.YCoord + slider.Height)) {
            if (XCoord < slider.XCoord) {
               /*** Slide left ***/
               if (XCoord >= Self->StartMargin) {
                  move.XChange = -move.XChange;
                  ActionMsg(AC_Move, Self->SliderID, &move);
               }
            }
            else if (XCoord > slider.XCoord + slider.Width) {
               if (XCoord <= (Self->StartMargin + Self->BarSize)) {
                  ActionMsg(AC_Move, Self->SliderID, &move);
               }
            }
         }
      }
      else if (Self->Flags & SCF_VERTICAL) {

         if ((XCoord >= slider.XCoord) AND (XCoord <= slider.XCoord + slider.Width)) {

            update.Position = -1;

            dir = DIR_NONE;
            if ((YCoord >= Self->StartMargin) AND (YCoord < slider.YCoord)) {
               dir = DIR_NEGATIVE;
            }
            else if ((YCoord > slider.YCoord + slider.Height) AND (YCoord <= (Self->StartMargin + Self->BarSize))) {
               dir = DIR_POSITIVE;
            }

            if (Self->Flags & SCF_REVERSE) {
               if (dir IS DIR_NEGATIVE) dir = DIR_POSITIVE;
               else if (dir IS DIR_POSITIVE) dir = DIR_NEGATIVE;
            }

            if (dir IS DIR_NEGATIVE) {
               update.Position = check_position(Self, Self->Position - (Self->ViewSize * PAGESCROLL));

               if (Self->Flags & SCF_RELATIVE) {
                  scroll.XChange = 0;
                  scroll.YChange = -ABS(Self->Position - update.Position);
                  scroll.ZChange = 0;
                  ActionMsg(AC_Scroll, Self->ObjectID, &scroll);
               }
               else {
                  scrollto.XCoord = 0;
                  scrollto.YCoord = update.Position;
                  scrollto.ZCoord = 0;
                  scrollto.Flags  = STF_Y;
                  ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
               }
            }
            else if (dir IS DIR_POSITIVE) {
               update.Position = check_position(Self, Self->Position + (Self->ViewSize * PAGESCROLL));

               if (Self->Flags & SCF_RELATIVE) {
                  scroll.XChange = 0;
                  scroll.YChange = ABS(Self->Position - update.Position);
                  scroll.ZChange = 0;
                  ActionMsg(AC_Scroll, Self->ObjectID, &scroll);
               }
               else {
                  scrollto.XCoord = 0;
                  scrollto.YCoord = update.Position;
                  scrollto.ZCoord = 0;
                  scrollto.Flags  = STF_Y;
                  ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
               }
            }

            if (update.Position != -1) {
               update.ViewSize = Self->ViewSize;
               update.PageSize = Self->PageSize;
               update.Unit     = Self->Unit;
               SCROLL_UpdateScroll(Self, &update);
            }
         }
      }
   }
   else {
      LONG change;

      for (i=0; i < ARRAYSIZE(Self->Buttons); i++) {
         if (Self->Buttons[i].ButtonID != NotifyID) continue;

         if (Self->Unit < 1) {
            if (Self->Flags & SCF_HORIZONTAL) {
               change = slider.Width;
               if (Self->Buttons[i].Direction IS SD_NEGATIVE) change = -change;
            }
            else {
               change = slider.Height;
               if (Self->Buttons[i].Direction IS SD_NEGATIVE) change = -change;
            }

            if (Self->Axis IS AXIS_X) move.XChange = change;
            else move.XChange = 0;
            if (Self->Axis IS AXIS_Y) move.YChange = change;
            else move.YChange = 0;
            if (Self->Axis IS AXIS_Z) move.ZChange = change;
            else move.ZChange = 0;

            ActionMsg(AC_Move, Self->SliderID, &move);
         }
         else {
            if (Self->Buttons[i].Direction IS SD_NEGATIVE) update.Position = check_position(Self, Self->Position - Self->Unit);
            else update.Position = check_position(Self, Self->Position + Self->Unit);

            if (update.Position != Self->Position) {
                  update.ViewSize = Self->ViewSize;
                  update.PageSize = Self->PageSize;
                  update.Unit     = Self->Unit;
                  SCROLL_UpdateScroll(Self, &update);

                  if (Self->Flags & SCF_RELATIVE) {
                     if (Self->Buttons[i].Direction IS SD_NEGATIVE) change = -Self->Unit;
                     else change = Self->Unit;

                     if (Self->Axis IS AXIS_X) scroll.XChange = change;
                     else scroll.XChange = 0;
                     if (Self->Axis IS AXIS_Y) scroll.YChange = change;
                     else scroll.YChange = 0;
                     if (Self->Axis IS AXIS_Z) scroll.ZChange = change;
                     else scroll.ZChange = 0;
                     ActionMsg(AC_Scroll, Self->ObjectID, &scroll);
               }
               else {
                  if (Self->Axis IS AXIS_X) { scrollto.XCoord = Self->Position; scrollto.Flags = STF_X; }
                  else if (Self->Axis IS AXIS_Y) { scrollto.YCoord = Self->Position; scrollto.Flags = STF_Y; }
                  else if (Self->Axis IS AXIS_Z) { scrollto.ZCoord = Self->Position; scrollto.Flags = STF_Z; }
                  ActionMsg(AC_ScrollToPoint, Self->ObjectID, &scrollto);
               }
            }
         }
      }
   }

#ifdef DEBUG
   StepBack();
#endif

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    AddScrollButton
Synonym: AddButton
Short:   Registers a new button with the scroll object.

###DESCRIPTION###
<p>If you are using the scroll object to create a complete scrollbar, you
may want to create buttons at either end of the bar. You will need
to register new buttons with the scroll object, which you may do by calling the
AddButton method.</p>

<p>A button is typically created from the Button or Render classes, but you
may use any object type that belongs to a class supporting the UserClick and
UserClickRepeat actions.  To register the button you need to provide the
AddScrollButton method with the button's unique ID.  The Direction
represented by the button must also be specified as SD_NEGATIVE or
SD_POSITIVE.</p>

###INPUT###
OBJECTID/Button: Reference to a Button, Render object or any class type that supports UserClick and UserClickRepeat actions..
LONG/Direction: The direction that the button represents - SD_NEGATIVE or SD_POSITIVE.

###RESULT###
ERR_Okay: The button was registered successfully.
ERR_Args: The required arguments were not specified.

###END###

*****************************************************************************/

static ERROR SCROLL_AddButton(struct Scroll *Self, struct mtAddScrollButton *Args)
{
   OBJECTPTR object;
   WORD i;

   if ((!Args) OR (!Args->ButtonID) OR (!Args->Direction)) return(ObjectError(Self, MT_AddScrollButton, ERR_Args));

   DPrintF("~AddButton()","[Scroll:%d] %d", Self->Head.UniqueID, Args->ButtonID);

   for (i=0; i < ARRAYSIZE(Self->Buttons); i++) {
      if (Self->Buttons[i].Direction IS Args->Direction) break;
      if (!Self->Buttons[i].ButtonID) break;
   }

   if (i >= ARRAYSIZE(Self->Buttons)) return(ObjectError(Self, MT_AddScrollButton, ERR_ArrayFull));

   if (Self->Buttons[i].ButtonID) {
      if (AccessObject(Self->Buttons[i].ButtonID, 5000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, Self->Head.UniqueID);
         ReleaseObject(object);
      }
   }

   if (AccessObject(Args->ButtonID, 5000, &object) IS ERR_Okay) {
      SubscribeActionTags(object, Self, AC_UserClickRelease, AC_UserClickRepeat, AC_UserClick, TAGEND);
      ReleaseObject(object);

      Self->Buttons[i].ButtonID = Args->ButtonID;

      if (Args->Direction IS 3) Self->Buttons[i].Direction = SD_NEGATIVE; /* Backwards compatible with SD_LEFT */
      else if (Args->Direction IS 4) Self->Buttons[i].Direction = SD_POSITIVE; /* Backwards compatible with SD_RIGHT */
      else Self->Buttons[i].Direction = Args->Direction;
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Scroll: Init
*/

static ERROR SCROLL_Init(struct Scroll *Self, APTR Void)
{
   struct Scroll *intersect;
   OBJECTPTR drawable, object, bar;
   LONG offset, size, visible;

   /*** Find the drawable object that we are associated with ***/

   if (!Self->SliderID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->SliderID = OwnerID;
   }

   if (!Self->ObjectID) DPrintF("@Init:","[Scroll:%d] Warning: The Object field is not set.", Self->Head.UniqueID);

   /*** Is the scrollbar horizontal or vertical? ***/

   if (!(Self->Flags & (SCF_HORIZONTAL|SCF_VERTICAL))) {
      Self->Flags |= SCF_VERTICAL;
   }

   /* Monitor the scroll container for movement, and the Slider's drawable
   ** container for Resize actions.
   */

   if (AccessObject(Self->SliderID, 5000, &drawable) IS ERR_Okay) {
      OBJECTPTR bar;

      Self->ScrollbarID = GetOwner(drawable);

      SubscribeAction(drawable, AC_MoveToPoint, Self);

      GetFields(drawable, FID_XCoord|TLONG, &Self->SliderX,
                          FID_YCoord|TLONG, &Self->SliderY,
                          TAGEND);

      /* Calculate the inside bar height/width and subscribe to
      ** the Scrollbar's Resize action.
      */

      if (AccessObject(Self->ScrollbarID, 5000, &bar) IS ERR_Okay) {
         SubscribeActionTags(bar, Self,
            AC_Redimension,
            AC_UserClick,
            AC_UserClickRepeat,
            TAGEND);

         if (Self->Flags & SCF_VERTICAL) {
            GetField(bar, FID_Height, FT_LONG, &Self->BarSize);

            if (Self->StartMargin IS -1) GetField(drawable, FID_TopLimit, FT_LONG, &Self->StartMargin);
            if (Self->EndMargin IS -1) GetField(drawable, FID_BottomLimit, FT_LONG, &Self->EndMargin);
         }
         else if (Self->Flags & SCF_HORIZONTAL) {
            GetField(bar, FID_Width, FT_LONG, &Self->BarSize);

            if (Self->StartMargin IS -1) GetField(drawable, FID_LeftLimit, FT_LONG, &Self->StartMargin);
            if (Self->EndMargin IS -1) GetField(drawable, FID_RightLimit, FT_LONG, &Self->EndMargin);
         }

         if (Self->ViewSize < 1) Self->ViewSize = Self->BarSize - Self->ObscuredView;

         Self->BarSize -= (Self->StartMargin + Self->EndMargin);

         ReleaseObject(bar);
      }

      ReleaseObject(drawable);
   }

   /* Subscribe to the Scroll action of the target object.  This allows us to adjust
   ** the sliders in the event that somebody scrolls the target object without
   ** informing us directly.
   */

   if (AccessObject(Self->ObjectID, 5000, &object) IS ERR_Okay) {
      SubscribeActionTags(object, Self, AC_Scroll, TAGEND);
      ReleaseObject(object);
   }

   /* If a drawable is to be monitored for scroll commands (e.g. from the
   ** mouse wheel) we will subscribe to it here.
   */

   //if (Self->MonitorID IS Self->ObjectID) Self->MonitorID = NULL;

   if ((Self->MonitorID) AND (Self->MonitorID != Self->ObjectID)) {
      if (AccessObject(Self->MonitorID, 5000, (OBJECTPTR *)&object) IS ERR_Okay) {
         SubscribeActionTags(object, Self, AC_Scroll, TAGEND);
         ReleaseObject(object);
      }
      else Self->MonitorID = NULL;
   }

   /* If an intersecting scrollbar has been specified,
   ** subscribe to its drawable's Hide and Show actions.
   */

   if ((Self->IntersectID) AND (Self->IntersectDrawable)) {
      if (AccessObject(Self->IntersectID, 5000, (OBJECTPTR *)&intersect) IS ERR_Okay) {
         if (AccessObject(Self->IntersectDrawable, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
            SubscribeActionTags(drawable, Self, AC_Hide, AC_Show, TAGEND);

            SetFields(intersect, FID_Intersect|TLONG, Self->Head.UniqueID, TAGEND);

            /* Position ourselves according to whether or not the intersecting
            ** scrollbar is visible.
            */

            GetFields(drawable, FID_Visible|TLONG, &visible, TAGEND);

            if (AccessObject(Self->ScrollbarID, 5000, &bar) IS ERR_Okay) {
               if (visible) {
                  if (Self->Flags & SCF_HORIZONTAL) {
                     GetFields(drawable, FID_XOffset|TLONG, &offset, FID_Width|TLONG, &size, TAGEND);
                     SetFields(bar, FID_XOffset|TLONG, offset + size, TAGEND);
                  }
                  else if (Self->Flags & SCF_VERTICAL) {
                     GetFields(drawable, FID_YOffset|TLONG, &offset, FID_Height|TLONG, &size, TAGEND);
                     SetFields(bar, FID_YOffset|TLONG, offset + size, TAGEND);
                  }

                  DPrintF("Init:","[Scroll] Intersection bar is visible, shrunk to offset %d.", offset - size);
               }
               else {
                  if (Self->Flags & SCF_HORIZONTAL) {
                     GetFields(drawable, FID_XOffset|TLONG, &offset, TAGEND);
                     SetFields(bar, FID_XOffset|TLONG, offset, TAGEND);
                  }
                  else if (Self->Flags & SCF_VERTICAL) {
                     GetFields(drawable, FID_YOffset|TLONG, &offset, TAGEND);
                     SetFields(bar, FID_YOffset|TLONG, offset, TAGEND);
                  }
                  DPrintF("Init:","[Scroll] Intersection bar is invsiible, expanded to offset %d.", offset);
               }
               ReleaseObject(bar);
            }
            ReleaseObject(drawable);
         }
         else Self->IntersectID = NULL;

         ReleaseObject(intersect);
      }
      else Self->IntersectID = NULL;
   }

   /* If the PageSize and ViewSize values have been specified, set up
   ** the scrolling area to reflect the settings.
   */

   if ((Self->PageSize) AND (Self->ViewSize)) {
      struct mtUpdateScroll scroll;

      DPrintF("Init:","[Scroll:%d] Preset PageSize %d, ViewSize %d and Position %.2f", Self->Head.UniqueID, Self->PageSize, Self->ViewSize, Self->Position);

      scroll.PageSize = Self->PageSize;
      scroll.ViewSize = Self->ViewSize;
      scroll.Position = Self->Position;
      scroll.Unit     = Self->Unit;
      SCROLL_UpdateScroll(Self, &scroll);
   }

   if (!Self->Axis) {
      if (Self->Flags & SCF_HORIZONTAL) Self->Axis = AXIS_X;
      else Self->Axis = AXIS_Y;
   }

   DPrintF("Init:","[Scroll:%d] Object: %d, Slider: %d, Scrollbar: %d", Self->Head.UniqueID, Self->ObjectID, Self->SliderID, Self->ScrollbarID);

   return(ERR_Okay);
}

/*****************************************************************************
** Scroll: NewObject
*/

static ERROR SCROLL_NewObject(struct Scroll *Self, APTR Void)
{
   Self->Unit        = 1;
   Self->PrevCoord   = -1;
   Self->StartMargin = -1;
   Self->EndMargin   = -1;
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  UpdateScroll
Short: Updates the dimensions of a scroll object's slider.

###DESCRIPTION###
<p>After a scroll object has been initialised, the only way to make
changes to the PageSize, Position or ViewSize is to use the UpdateScroll()
method.  When you call this method the scroll object will update the size
and position of the slider so that it correctly reflects the new values.</p>

<p>This method is typically used by programs and classes that have scrollbar
awareness.</p>

###INPUT###
LONG/PageSize: The size of the page.  Set to zero for no change.
LONG/ViewSize: The size of the view of the page.  Set to zero for no change.
LONG/Position: The current position within the page.
LONG/Unit:     The unit size to use for micro-scrolling.  Set to zero for no change, or -1 to enable jump scrolling.
###END###

*****************************************************************************/

static ERROR SCROLL_UpdateScroll(struct Scroll *Self, struct mtUpdateScroll *Args)
{
   struct acRedimension redimension;
   DOUBLE scale;
   LONG offset, pos, minsize;

   if (Args) {
      Self->Position = Args->Position;
      if (Args->PageSize > 0) Self->PageSize = Args->PageSize;
      if (Args->ViewSize > 0) Self->ViewSize = Args->ViewSize;
      if (Args->Unit) Self->Unit = Args->Unit;
   }

   #ifdef DEBUG
      DPrintF("UpdateScroll()","[Scroll:%d] Position: %d, Page: %d, View: %d, Unit: %d", Self->Head.UniqueID, Self->Position, Self->PageSize, Self->ViewSize, Self->Unit);
   #endif

   if ((Self->PageSize < 0) OR (Self->ViewSize <= 0)) {
      DPrintF("@UpdateScroll:","Illegal pagesize/viewsize settings of %d/%d respectively.", Self->PageSize, Self->ViewSize);
      return(ERR_Okay);
   }

   if (!Self->PageSize) {
      Self->Position = 0;
      Self->PageSize = 1; /* Set to 1 in order to prevent division by zero errors */
   }

   Self->Position = check_position(Self, Self->Position);

   pos = Self->Position;
   if (Self->Flags & SCF_REVERSE) pos = Self->PageSize - pos;
   if (pos < 0) pos = 0;

   if (!(Self->Flags & SCF_FIXED)) {
      /*** Proportional slider ***/
      Self->SliderSize = Self->BarSize * Self->ViewSize / Self->PageSize;
   }

   if ((Self->Flags & (SCF_SLIDER|SCF_FIXED))) minsize = 11;
   else minsize = 20;

   if (Self->SliderSize < minsize) {
      /*** This routine is for a fixed slider ***/

      Self->SliderSize = minsize;

      scale = (DOUBLE)(Self->PageSize - Self->ViewSize) / (DOUBLE)(Self->BarSize - minsize);
      offset = pos / scale;

      /* Do not allow the slider size to exceed the maximum amount
      ** of movement space available to the slider.
      */

      if ((Self->SliderSize + offset) > Self->BarSize) {
         offset = Self->BarSize - Self->SliderSize;
         if (offset < 0) offset = 0;
         if ((Self->SliderSize + offset) > Self->BarSize) {
            Self->SliderSize = Self->BarSize;
         }
      }
   }
   else {
      if (Self->Flags & SCF_SLIDER) {
         /*** This routine is for a proportional slider ***/
         offset = pos * (Self->SliderSize - Self->BarSize) / Self->PageSize;
      }
      else {
         /*** This routine is for a proportional scrollbar ***/

         if ((pos + Self->ViewSize) IS Self->PageSize) offset = Self->BarSize - Self->SliderSize;
         else offset = pos * (Self->BarSize) / Self->PageSize;
      }

      /* Do not allow the slider size to exceed the maximum amount
      ** of movement space available to the slider.
      */

      if ((Self->SliderSize + offset) > Self->BarSize) {
         Self->SliderSize = Self->BarSize - offset;
      }
   }

   /*** Set the values ***/

   if (offset < 0) {
      #ifdef DEBUG
         DPrintF("@UpdateScroll:","Calculated illegal slider offset of %d.", offset);
      #endif
      offset = -offset;
   }

   if (Self->Flags & SCF_VERTICAL) {
      redimension.XCoord = Self->SliderX;
      redimension.YCoord = Self->StartMargin + offset;
      redimension.ZCoord = 0;
      redimension.Width  = 0;
      redimension.Height = Self->SliderSize;
      redimension.Depth  = 0;
      ActionMsg(AC_Redimension, Self->SliderID, &redimension);

      if (Self->PrevCoord IS -1) Self->PrevCoord = Self->StartMargin + offset;
   }
   else if (Self->Flags & SCF_HORIZONTAL) {
      redimension.XCoord = Self->StartMargin + offset;
      redimension.YCoord = Self->SliderY;
      redimension.ZCoord = 0;
      redimension.Width  = Self->SliderSize;
      redimension.Height = 0;
      redimension.Depth  = 0;
      ActionMsg(AC_Redimension, Self->SliderID, &redimension);

      if (Self->PrevCoord IS -1) Self->PrevCoord = Self->StartMargin + offset;
   }

   /* This code decides whether it is necessary to see the
   ** scrollbar or not, according to its values.
   */

   if (Self->Flags & SCF_AUTOHIDE) {
      if (Self->PageSize <= 1) {
         acHideID(Self->ScrollbarID);
      }
      else if ((offset IS 0) AND ((Self->PageSize <= Self->ViewSize) OR (Self->SliderSize >= Self->BarSize))) {
         acHideID(Self->ScrollbarID);
      }
      else acShowID(Self->ScrollbarID);
   }

   /* If an object field is linked to us, we must always ensure that it is told of
   ** the current position.
   */

   if ((Self->FieldID) AND (Self->ObjectID)) {
      OBJECTPTR object;
      if (AccessObject(Self->ObjectID, 5000, &object) IS ERR_Okay) {
         SetField(object, Self->FieldID, FT_LONG, &Self->Position);
         ReleaseObject(object);
      }
   }

   #ifdef DEBUG
      DPrintF("UpdateScroll:","Position: %d, Page: %d, View: %d, Slider: %d/%d", Self->Position, Self->PageSize, Self->ViewSize, Self->SliderSize, Self->BarSize);
   #endif

   set_position(Self, Self->Position);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Axis
Short:  The axis that the scroll object represents can be defined here.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>When a scroll slider is moved, scroll messages are sent for one axis only
- either X, Y, or Z.  You need to define the axis here using one of the
constants AXIS_X, AXIS_Y or AXIS_Z.  The axis does not necessarily have to
match the orientation of the scrollbar.  For instance, a horizontal
scrollbar can send vertical scroll messages if you use AXIS_Y.</p>

###END###

*****************************************************************************/

static ERROR SET_Axis(struct Scroll *Self, LONG *Value)
{
   if (*Value IS AXIS_X) Self->Axis = AXIS_X;
   else if (*Value IS AXIS_Y) Self->Axis = AXIS_Y;
   else if (*Value IS AXIS_Z) {
      Self->Axis = AXIS_Z;
      Self->Flags |= SCF_SLIDER;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Defines the 'slider' that represents the scroll object.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In order to work correctly, a scroll object must be related to a drawable
that acts as a slider.  In most cases you should initialise the scroll object
to the drawable that will act as the slider.  If you do not do this then you
will have to tell the scroll object exactly what drawable is acting as the
slider, and thus will need to refer to the slider object via the Drawable
field.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Field
Short:  Reference to an object field that should be updated with the scroll value.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>If you would like to write the position of a scroll object to a field
belonging to another object (refer to the Object field), you can make reference
to the field name here.  By doing this, whenever the scroll object updates its
internal position value, it will also write that value to the referenced field
name.</p>

###END###

*****************************************************************************/

static ERROR SET_Field(struct Scroll *Self, STRING Value)
{
   if (Value) {
      ResolveFields(Value, &Self->FieldID, TAGEND);
   }
   else return(ERR_Failed);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Special scroll flags.
Type:   LONG
Prefix: SCF
Status: Read/Set

###DESCRIPTION###
<p>Flags supported by the scroll object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">AUTOHIDE</td><td>If this flag is on then the scroll object will automatically hide the container of the slider when the size of the view exceeds or is equal to the size of the page.</td></tr>
<tr><td width="10%">HORIZONTAL&nbsp;&nbsp;</td><td>If the scroll object's slider will represent horizontal movement, set this flag.</td></tr>
<tr><td width="10%">VERTICAL</td><td>If the scroll object's slider will represent vertical movement, set this flag.</td></tr>
<tr><td width="10%">RELATIVE</td><td>Forces relative scrolling through the Scroll action rather than the default ScrollToPoint.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Intersect
Short:  This field is used for managing intersecting scrollbars.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>When two scroll objects are used to create opposing scrollbars (e.g.
horizontal and vertical bars) then you should set the Intersect field if
there is an overlap between the two.  The purpose of this is to keep the
scrollbars neatly arranged if one of them disappears (which will occur if
the viewable area is larger than the size of the scrollable page).</p>

<p>The Intersect field must be set to a valid scroll object that represents
the opposing scrollbar.  The opposing scroll object does not need to have
its Intersect field set.</p>

<p>If you use the ScrollBar class, intersections are managed
automatically.</p>

###END###

*****************************************************************************/

static ERROR SET_Intersect(struct Scroll *Self, OBJECTID *ObjectID)
{
   struct Scroll *intersect;
   OBJECTPTR drawable;

   if ((Self->IntersectID = *ObjectID)) {
      if (GetClassID(Self->IntersectID) != ID_SCROLL) {
         Self->IntersectID = NULL;
         DPrintF("@Scroll:","The Intersect field can only be set with valid Scroll objects.");
         return(ERR_Failed);
      }

      if (AccessObject(Self->IntersectID, 5000, (OBJECTPTR *)&intersect) IS ERR_Okay) {
         Self->IntersectDrawable = intersect->ScrollbarID;
         ReleaseObject(intersect);
      }
      else {
         Self->IntersectID = NULL;
         return(ERR_AccessObject);
      }
   }

   /* If we have been initialised already, then this is a post-intersection
   ** setting.  In this mode we are not required to adjust our scrollbar
   ** position, but we do need to make adjustments to the available viewing
   ** area in the event that the intersecting scrollbar is obscuring the
   ** scrollable page.
   */

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      Self->PostIntersect = TRUE;
      if (AccessObject(Self->IntersectDrawable, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
         SubscribeActionTags(drawable, Self, AC_Hide, AC_Show, TAGEND);
         ReleaseObject(drawable);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Monitor
Short:  Objects can be monitored for scroll requests by setting this field.
Type:   OBJECTID
Status: Set

###DESCRIPTION###
<p>To monitor an object for scroll requests, set this field to a valid
object ID.  This feature is often used to support the wheel mouse, for
instances where the mouse is positioned over a drawable area and the wheel
is used.  The use of the wheel will cause Scroll messages to be sent from
the mouse to the underlying drawable.  By setting this field to the drawable
area that is being scrolled, the scrollbar can receive and respond to the
scroll messages.</p>

###END###

*****************************************************************************/

static ERROR SET_Monitor(struct Scroll *Self, OBJECTID *Value)
{
   OBJECTPTR object;

   if (Self->MonitorID IS *Value) return(ERR_Okay);

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      if (Self->MonitorID IS Self->ObjectID) {
         /*** Do nothing because we will already have subscribed to the Scroll action ***/
      }
      else {
         if (Self->MonitorID) {
            if (AccessObject(Self->MonitorID, 5000, &object) IS ERR_Okay) {
               UnsubscribeAction(object, AC_Scroll, Self->Head.UniqueID);
               Self->MonitorID = NULL;
               ReleaseObject(object);
            }
         }

         if (!Value[0]) {
            Self->MonitorID = NULL;
         }
         else if (AccessObject(*Value, 5000, (OBJECTPTR *)&object) IS ERR_Okay) {
            SubscribeActionTags(object, Self, AC_Scroll, TAGEND);
            Self->MonitorID = *Value;
            ReleaseObject(object);
         }
         else return(ObjectError(Self, ERH_SetField, ERR_AccessObject));
      }
   }
   else Self->MonitorID = *Value;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Object
Short:  Refers to the object that will receive the scroll action.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>The Object field must refer to a foreign object that will receive Scroll
notification calls whenever the scroll object is moved.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   PageSize
Short:  Defines the size of the page that is to be scrolled.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The page size of the area that is being scrolled is declared through this
field.  The page size should almost always be larger than the view size,
because the page lies 'under' the view.  If the page is smaller than the
view, the scroll object will serve no purpose until the circumstances are
changed.</p>

###SEE ALSO###
Field: ViewSize

###END###

*****************************************************************************/

static ERROR SET_PageSize(struct Scroll *Self, LONG *Value)
{
   struct mtUpdateScroll update;

   Self->PageSize = *Value;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      update.ViewSize = Self->ViewSize;
      update.PageSize = Self->PageSize;
      update.Position = Self->Position;
      update.Unit     = Self->Unit;
      Action(MT_UpdateScroll, Self, &update);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Position
Short:  Reflects the current position of the page within the view.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>As the page moves within the view, the exact position of the page must
be recorded so that the scroll gadget is correctly positioned within the
scrollbar area.  When setting up a scroll object, you may need to set the
Position value if the page starts at a certain offset within the view.
Otherwise, you should leave the field at the default position of 0.</p>

###END###

*****************************************************************************/

static ERROR SET_Position(struct Scroll *Self, LONG *Value)
{
   struct mtUpdateScroll update;

   Self->Position = *Value;

   if (Self->Position > Self->PageSize) Self->Position = Self->PageSize;
   if (Self->Position < 0) Self->Position = 0;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      update.ViewSize = Self->ViewSize;
      update.PageSize = Self->PageSize;
      update.Position = Self->Position;
      update.Unit     = Self->Unit;
      Action(MT_UpdateScroll, Self, &update);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SliderSize
Short:  Can refer to a drawable acting as the 'right button'.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>The SliderSize field can be read if you need to know the size of the slider
represented by a scroll object.  You can also set this field during initialisation
if you would like a fixed slider to be used (this results in a non-proportional
scrollbar).</p>

###END###

*****************************************************************************/

static ERROR SET_SliderSize(struct Scroll *Self, LONG *Value)
{
   Self->SliderSize = *Value;
   Self->Flags |= SCF_FIXED;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Unit
Short:  Defines the amount of movement used when a scroll button is pressed.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you have attached buttons to a scroll object through the In, Out, Left,
Right, Up or Down fields then you may want to specify the amount of units to
use when a related button is pressed.  Example:  If the unit is set to 15
and the scroll position is currently set to 60, pressing a negative button
would change the position to 45.</p>

<p>The Unit value should normally be positive, but if you set it to 0 or
less then the Unit will be dynamically calculated to match the size of the
slider.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ViewSize
Short:  Defines the size of the view that contains the page.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The ViewSize defines the width or height of the area that is viewing the
page.  The view should almost always be smaller than the size of the page,
because the view sits above it.  There are legitimate exceptions however, so
this particular rule is not enforced.</p>

###SEE ALSO###
Field: PageSize

###END###

*****************************************************************************/

static ERROR SET_ViewSize(struct Scroll *Self, LONG *Value)
{
   struct mtUpdateScroll update;

   Self->ViewSize = *Value;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      update.ViewSize = Self->ViewSize;
      update.PageSize = Self->PageSize;
      update.Position = Self->Position;
      update.Unit     = Self->Unit;
      Action(MT_UpdateScroll, Self, &update);
   }

   return(ERR_Okay);
}

/*****************************************************************************

                      "We stab their bodies with daggers
                       We smash their skulls with clubs
                       We crush their bones
                       And gash their faces
                       Their necks are broken and severed
                       Their bowels are gored with spears
                       Yeah!  We hack off their limbs
                       And cast the phalluses
                       Into pits of fire"

                       - Nile

*****************************************************************************/
