/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      IconServer
Version:   1.0
Status:    Stable
ID:        ID_ICONSERVER
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2003-2004.  All rights reserved.
Keywords:  icons, filesystem

###DESCRIPTION###
<p>The IconServer class provides an icon management and image resizing service
for Athene.  It provides this service through the icons: assignment.  Icons
are requested from the server by accessing them through the virtual filesystem,
making access to the icon server 'invisible'.  Direct contact with the icon
server only needs to be made when requesting a new theme, or to change the
default icon size.</p>

<p>Icon files can be accessed from the icon server using the following file
path formats:</p>

<pre>
   icons:filename
   icons:category/filename
   icons:category/filename(size)
</pre>

<p>The second format, where the category and filename is specified is
recommended for normal usage.  The size of the icon can be requested by entering
a numeral in brackets at the end of the location.  For instance
'icons:programs/filemanager(64)' will generate a 64x64 pixel version of the
filemanager icon.  The minimum size for an icon is 16 pixels and the maximum
size is 256 pixels.</p>

###END###

*****************************************************************************/

#define __system__
#include <pandora/system/all.h>
#include <pandora/files/all.h>
#include <pandora/graphics/picture.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/functions/set.c>

#define VER_IconServer 1.0

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_IconServer, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003-2004.  All rights reserved.",
   "August 2004",
   "IconServer"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static OBJECTPTR modStrings = 0;
static OBJECTPTR modFileSystem = 0;
static OBJECTPTR IconServerClass = 0;

enum {
   CODE_CATEGORY=1,
   CODE_ICON,
   CODE_END
};

/*****************************************************************************
** IconServer class.
*/

struct IconServer {
   OBJECT_HEADER
   LONG   FixedSize;
   ERROR  (*ResolveIconAssign)(struct IconServer *, STRING, STRING, LONG);
   FLOAT  IconRatio;
   LONG   VolatileIcons;

   PRIVATE_FIELDS
   UBYTE prvTheme[60];
   APTR Database;
   MEMORYID DatabaseMID;
   LONG DataSize;
};

static ERROR ICON_AccessObject(struct IconServer *, APTR);
static ERROR ICON_Clear(struct IconServer *, APTR);
static ERROR ICON_Free(struct IconServer *, APTR);
static ERROR ICON_Init(struct IconServer *, APTR);
static ERROR ICON_NewObject(struct IconServer *, APTR);
static ERROR ICON_Refresh(struct IconServer *, APTR);
static ERROR ICON_ReleaseObject(struct IconServer *, APTR);

static ERROR GET_CacheSize(struct IconServer *, LONG *);
static ERROR GET_IconSize(struct IconServer *, LONG *);
static ERROR GET_Theme(struct IconServer *, STRING *);

static ERROR SET_IconSize(struct IconServer *, LONG *);
static ERROR SET_IconRatio(struct IconServer *, FLOAT *);
static ERROR SET_Theme(struct IconServer *, STRING);

static struct FieldArray IconServerFields[] = {
   { "FixedSize",     0, FDF_LONG|FDF_RW,    0, NULL, SET_IconSize }, /* When reading the FixedSize field, the preferred size is returned (which can be 0 if fixed sizing is off) */
   { "ResolveAssign", 0, FDF_POINTER|FDF_RW, 0, NULL, NULL },
   { "IconRatio",     0, FDF_FLOAT|FDF_RW,   0, NULL, SET_IconRatio },
   { "VolatileIcons", 0, FDF_LONG|FDF_RW,    0, NULL, NULL },
   /*** Virtual fields ***/
   { "CacheSize",     0, FDF_LONG|FDF_R,     0, GET_CacheSize, NULL },
   { "IconSize",      0, FDF_LONG|FDF_RW,    0, GET_IconSize, SET_IconSize },
   { "Theme",         0, FDF_STRING|FDF_RW,  0, GET_Theme, SET_Theme },
    END_FIELD
};

static struct ActionArray IconServerActions[] = {
   { AC_AccessObject,  ICON_AccessObject },
   { AC_Clear,         ICON_Clear },
   { AC_Free,          ICON_Free },
   { AC_Init,          ICON_Init },
   { AC_NewObject,     ICON_NewObject },
   { AC_Refresh,       ICON_Refresh },
   { AC_ReleaseObject, ICON_ReleaseObject },
   { NULL, NULL }
};

/****************************************************************************/

static ERROR LoadIconDatabase(struct IconServer *);
static ERROR ResolveIconAssign(struct IconServer *, STRING, STRING, LONG);
static void WriteIconCategory(struct IconServer *, OBJECTPTR, STRING);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Create the icon server class ***/

   if (CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&IconServerClass, NULL,
      FID_BaseClassID|TLONG, ID_ICONSERVER,
      FID_Version|TFLOAT,    VER_IconServer,
      FID_Name|TSTRING,      "IconServer",
      FID_Category|TLONG,    CCF_SYSTEM,
      FID_Actions|TPTR,      IconServerActions,
      FID_Fields|TPTR,       IconServerFields,
      FID_Flags|TLONG,       CLF_SHAREDONLY,
      FID_Size|TLONG,        sizeof(struct IconServer),
      TAGEND) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));
   }

   /*** Add the icons: special assignment ***/

   SetAssign(AST_NAME, "icons", AST_PATH, ":SystemIcons", AST_FLAGS, ASSIGN_REPLACE|ASSIGN_HIDDEN, AST_ICON, "programs/iconthemes", TAGEND);

   return(ERR_Okay);
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (IconServerClass) { acFree(IconServerClass); IconServerClass = NULL; }
   if (modStrings)      { acFree(modStrings);      modStrings      = NULL; }
   if (modFileSystem)   { acFree(modFileSystem);   modFileSystem   = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** IconServer: AccessObject
*/

static ERROR ICON_AccessObject(struct IconServer *Self, APTR Void)
{
   if (Self->DatabaseMID) {
      if (AccessMemory(Self->DatabaseMID, MEM_READWRITE, (APTR *)&Self->Database) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessMemory));
      }
   }

   Self->ResolveIconAssign = &ResolveIconAssign;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Clear
Short: Removes all cached images generated by the icon server.
###END###

*****************************************************************************/

static ERROR ICON_Clear(struct IconServer *Self, APTR Void)
{
   UBYTE buffer[256];
   struct FileInfo *list, *scan;

   DPrintF("~Clear()","[IconServer]");

   StrFormat(buffer, sizeof(buffer), "pictures:icons/%s/", Self->prvTheme);

   if (ReadDirectory(buffer, RDF_DIRECTORIES, &list) IS ERR_Okay) {
      for (scan=list; scan; scan=scan->Next) {
         StrFormat(buffer, sizeof(buffer), "user:temp/iconcache/%s/%s/", Self->prvTheme, scan->Name);

         DeleteFile(buffer);
      }

      FreeMemory(list);
   }
   else DPrintF("Clear:","[IconServer] No image directories in theme location \"%s\".", buffer);

   StepBack();

   return(ERR_Okay);
}

/*****************************************************************************
** IconServer: Free()
*/

static ERROR ICON_Free(struct IconServer *Self, APTR Void)
{
   if (Self->DatabaseMID) {
      if (Self->Database) {
         ReleaseMemory(Self->Database);
         Self->Database = NULL;
      }
      FreeMemoryID(Self->DatabaseMID);
      Self->DatabaseMID = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** IconServer: Init()
*/

static ERROR ICON_Init(struct IconServer *Self, APTR Void)
{
   UBYTE buffer[100];
   LONG type;

   /* Check that the theme directory exists.  Switch to the default
   ** directory if it does not.
   */

   DPrintF("Init:","[IconServer] Checking for theme \"%s\".", Self->prvTheme);

   StrFormat(buffer, sizeof(buffer), "pictures:icons/%s/", Self->prvTheme);
   if ((AnalyseLocation(buffer, &type) != ERR_Okay) OR (type != LOC_DIRECTORY)) {
      DPrintF("Init:","[IconServer] Bad theme \"%s\", switching to default.", Self->prvTheme);
      StrCopy("Default", Self->prvTheme, sizeof(Self->prvTheme));
   }

   /* Open the iconserver.data file from the user's icon theme directory.
   ** If the iconserver.data file does not exist, we'll rebuild the file by scanning the icon
   ** directories.
   */
/*
   if (LoadIconDatabase(Self) IS ERR_Okay) {
      return(ERR_Okay);
   }
*/
   if (acRefresh(Self) IS ERR_Okay) {
      return(ERR_Okay);
   }

   return(ERR_Okay);
}

/****************************************************************************
** IconServer: NewObject
*/

static ERROR ICON_NewObject(struct IconServer *Self, APTR Void)
{
   StrCopy("Default", Self->prvTheme, sizeof(Self->prvTheme));
   Self->FixedSize = 0;
   Self->IconRatio = 7.0;
   Self->ResolveIconAssign = &ResolveIconAssign;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Refresh
Short: Regenerates the icon database.
###END###

*****************************************************************************/

static ERROR ICON_Refresh(struct IconServer *Self, APTR Void)
{
   struct acWrite write;
   struct FileInfo *scan, *list;
   OBJECTPTR output;
   UBYTE buffer[256], *data;
   WORD len;

   /* Regenerate the icon database.  This involves scanning each directory in the
   ** theme location and querying them through the Picture class (so that we can
   ** obtain size information).
   **
   ** The database stores icons in this format:
   **
   ** <LONG:Code> <LONG:Length> <UBYTE[]:Category>
   **   <LONG:Code> <LONG:Length> <UBYTE[]:IconFile>
   **   <LONG:Code> <LONG:Length> <UBYTE[]:IconFile>
   **   ...
   ** <LONG:Code> <LONG:Length> <UBYTE[]:Category>
   **   <LONG:Code> <LONG:Length> <UBYTE[]:IconFile>
   **   <LONG:Code> <LONG:Length> <UBYTE[]:IconFile>
   **   ...
   */

   DPrintF("~Refresh:","[IconServer] Regenerating the icon database...");

   StrFormat(buffer, sizeof(buffer), "user:temp/iconcache/%s/", Self->prvTheme);
   MakeDirectory(buffer, PERMIT_READ|PERMIT_WRITE);

   StrFormat(buffer, sizeof(buffer), "user:temp/iconcache/%s/iconserver.data", Self->prvTheme);

   if (CreateObject(ID_FILE, NF_CHILD, &output, NULL,
      FID_Location|TSTRING,  buffer,
      FID_Flags|TLONG,       FL_NEW|FL_WRITE,
      FID_Permissions|TLONG, PERMIT_READ|PERMIT_WRITE,
      TAGEND) IS ERR_Okay) {

      StrFormat(buffer, sizeof(buffer), "pictures:icons/%s/", Self->prvTheme);

      if (ReadDirectory(buffer, RDF_DIRECTORIES, &list) IS ERR_Okay) {
         for (scan=list; scan; scan=scan->Next) {
            data = buffer;
            ((LONG *)data)[0] = CODE_CATEGORY;
            data += sizeof(LONG);

            for (len=0; scan->Name[len]; len++);
            len++;
            len = (len + 3) & (~3);

            ((LONG *)data)[0] = sizeof(LONG) + sizeof(LONG) + len;
            data += sizeof(LONG);
            data += StrCopy(scan->Name, data, COPY_ALL);
            *data++ = 0; *data++ = 0; *data++ = 0; *data++ = 0;

            write.Buffer = buffer;
            write.Length = sizeof(LONG) + sizeof(LONG) + len;
            Action(AC_Write, output, &write);

            WriteIconCategory(Self, output, scan->Name);
         }

         /*** Write the end-of-database marker ***/

         data = buffer;
         ((LONG *)data)[0] = CODE_END;
         write.Buffer = buffer;
         write.Length = sizeof(LONG);
         Action(AC_Write, output, &write);

         FreeMemory(list);
      }
      else DPrintF("Refresh:","No image directories in theme location \"%s\".", buffer);

      acFree(output);
   }

   if (LoadIconDatabase(Self) != ERR_Okay) {
      StepBack();
      return(ERR_Failed);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Icon: ReleaseObject
*/

static ERROR ICON_ReleaseObject(struct IconServer *Self, APTR Void)
{
   if (Self->Database) { ReleaseMemory(Self->Database); Self->Database = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Cache
Short:  Switch off the cache by writing a FALSE value to this field.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>By default, all generated icons are cached on the hard drive for high speed
loading whenever an icon is requested at a previously generated size.  This
can potentially be a problem when developing new icon sets or adjusting the
icon database, as you may want to regenerate each icon from scratch.  Turning
off the cache will typically fix such problems.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   CacheSize
Short:  Indicates the total size of the icon cache in bytes.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>This field can be read at any time to determine the total number of bytes
in use by the icon cache.  The result value is generated dynamically through
analysis of the icon directory structure.  Only the current theme is
included in the result value.</p>

###END###

*****************************************************************************/

static ERROR GET_CacheSize(struct IconServer *Self, LONG *Value)
{
   struct FileInfo *dirlist, *filelist, *filescan, *dirscan;
   UBYTE buffer[256];

   DPrintF("~GetField:","[IconServer.CacheSize]");

   StrFormat(buffer, sizeof(buffer), "pictures:icons/%s/", Self->prvTheme);

   *Value = 0;
   if (ReadDirectory(buffer, RDF_DIRECTORY, &dirlist) IS ERR_Okay) {
      for (dirscan=dirlist; dirscan; dirscan=dirscan->Next) {
         StrFormat(buffer, sizeof(buffer), "user:temp/iconcache/%s/%s/", Self->prvTheme, dirscan->Name);

         if (ReadDirectory(buffer, RDF_FILE|RDF_SIZE, &filelist) IS ERR_Okay) {
            for (filescan=filelist; filescan; filescan=filescan->Next) *Value += filescan->Size;
            FreeMemory(filelist);
         }
      }
      FreeMemory(dirlist);
   }

   StepBack();

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   IconSize
Short:  The default size of each icon generated by the server can be changed here.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>By default, the iconserver will generate icons at a size of 32 pixels when
no specific size has been requested.  You may set this field to increase or
decrease the default icon size.  The minimum and maximum values are 16 and 256
pixels respectively.</p>

###END###

*****************************************************************************/

static ERROR GET_IconSize(struct IconServer *Self, LONG *Value)
{
   OBJECTID displayid;
   OBJECTPTR object;
   LONG width, height;

   if (Self->FixedSize >= 16) {
      *Value = Self->FixedSize;
   }
   else {
      if (Self->IconRatio < 2) Self->IconRatio = 2;

      if (FastFindObject("SystemDisplay", NULL, &displayid, 1, NULL) IS ERR_Okay) {
         if (AccessObject(displayid, 3000, &object) IS ERR_Okay) {
            GetFields(object, FID_Width|TLONG,  &width,
                              FID_Height|TLONG, &height,
                              TAGEND);
            if (width < height) *Value = ((FLOAT)width) * Self->IconRatio / 100.0;
            else *Value = ((FLOAT)height) * Self->IconRatio / 100.0;

            if (*Value < 16) *Value = 16;

            ReleaseObject(object);
         }
         else *Value = 16;
      }
      else *Value = 16;
   }

   return(ERR_Okay);
}

static ERROR SET_IconSize(struct IconServer *Self, LONG *Value)
{
   if (!Value[0]) {
      /*** Turn off fixed size, make sure ratios are enabled ***/
      Self->FixedSize = 0;
      if (Self->IconRatio < 2) Self->IconRatio = 2;
   }
   else if (*Value >= 16) {
      Self->FixedSize = *Value;
      if (Self->FixedSize < 16) Self->FixedSize = 16;
      else if (Self->FixedSize > 256) Self->FixedSize = 256;
  }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   IconRatio
Short:  Setting this field enables dynamic icon sizing that is relative to the display size.
Type:   FLOAT
Status: Read/Set

###DESCRIPTION###
<p>The IconRatio field can be set if you wish to enable dynamic icon sizing.
This field is set as a percentage, typically around 7%, which would generate
icons at 7% of the current display's shortest dimension (usually the height).</p>

<p>The minimum ratio is 2% and the maximum is 15%.  When you write to this
field, fixed icon sizes will no longer be the default.  You can switch back to fixed
icon sizes if you write a value to the IconSize field.  Fixed size icons can
also be requested at any time on an individual basis by passing a size to the icons:
assignment when loading icons.</p>

<p>Regardless of the ratio setting, the maximum and minimum pixel size of any
icon is always 16 and 256 pixels respectively.</p>

###SEE ALSO###
Field: IconSize

###END###

*****************************************************************************/

static ERROR SET_IconRatio(struct IconServer *Self, FLOAT *Value)
{
   Self->IconRatio = *Value;

   if (Self->IconRatio < 2.0) Self->IconRatio = 2.0;
   else if (Self->IconRatio > 15.0) Self->IconRatio = 15.0;

   DPrintF("IconServer:","Icon size ratio changed to %.2f%.", Self->IconRatio);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Theme
Short:  The theme used by an iconserver object is set here.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The theme currently in use by the iconserver is set here.  You can alter
the theme at any time by writing the name of the new theme to this field.  The
name must be represented in the 'pictures:icons/' directory tree or the theme
will revert back to the default setting.  When you change the theme, the icon
database will be refreshed.  This may take some time depending on the number
of icons to process.</p>

###END###

*****************************************************************************/

static ERROR GET_Theme(struct IconServer *Self, STRING *Value)
{
   if (Self->prvTheme[0]) *Value = Self->prvTheme;
   else *Value = NULL;
   return(ERR_Okay);
}

static ERROR SET_Theme(struct IconServer *Self, STRING Value)
{
   UBYTE buffer[256];
   LONG type;

   if ((Value) AND (*Value)) {
      StrCopy(Value, Self->prvTheme, sizeof(Self->prvTheme));
   }
   else StrCopy("Default", Self->prvTheme, sizeof(Self->prvTheme));

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) return(ERR_Okay);

   /* Check that the theme directory exists.  Switch to the default
   ** directory if it does not.
   */

   StrFormat(buffer, sizeof(buffer), "pictures:icons/%s/", Self->prvTheme);
   if ((AnalyseLocation(buffer, &type) != ERR_Okay) OR (type != LOC_DIRECTORY)) {
      DPrintF("@Set.Theme:","[IconServer] The requested theme \"%s\" does not exist.", Self->prvTheme);
      StrCopy("Default", Self->prvTheme, sizeof(Self->prvTheme));
   }

   /* Open the iconserver.data file from the user's icon theme directory.
   ** If the iconserver.data file does not exist, we'll rebuild the file by scanning the icon
   ** directories.
   */

   if (LoadIconDatabase(Self) IS ERR_Okay) {
      return(ERR_Okay);
   }

   if (Action(AC_Refresh, Self, NULL) IS ERR_Okay) {
      return(ERR_Okay);
   }

   return(ERR_Failed);
}

/*****************************************************************************
** Internal: LoadIconDatabase()
*/

static ERROR LoadIconDatabase(struct IconServer *Self)
{
   UBYTE buffer[256];
   struct acRead read;
   OBJECTPTR file;

   /*** Free the existing database if we have loaded one ***/

   if (Self->DatabaseMID) {
      if (Self->Database) {
         ReleaseMemory(Self->Database);
         Self->Database = NULL;
      }
      FreeMemoryID(Self->DatabaseMID);
      Self->DatabaseMID = NULL;
   }

   StrFormat(buffer, sizeof(buffer), "user:temp/iconcache/%s/iconserver.data", Self->prvTheme);

   if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
      FID_Location|TSTRING, buffer,
      FID_Flags|TLONG,      FL_READ,
      TAGEND) IS ERR_Okay) {

      if ((GetField(file, FID_Size, FT_LONG, &Self->DataSize) IS ERR_Okay) AND (Self->DataSize > 0)) {

         if (AllocMemory(Self->DataSize, Self->Head.MemFlags, &Self->Database, &Self->DatabaseMID) IS ERR_Okay) {
            read.Buffer = Self->Database;
            read.Length = Self->DataSize;
            Action(AC_Read, file, &read);
         }
         else {
            acFree(file);
            return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
         }
      }
      else {
         acFree(file);
         return(ObjectError(Self, ERH_SetField, ERR_NoData));
      }

      acFree(file);
   }
   else return(ERR_Failed);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: WriteIconCategory()
*/

static void WriteIconCategory(struct IconServer *Self, OBJECTPTR Output, STRING Category)
{
   struct FileInfo *list, *scan;
   struct acWrite write;
   UBYTE buffer[256], *data;
   WORD j, len;

   StrFormat(buffer, sizeof(buffer), "pictures:icons/%s/%s/", Self->prvTheme, Category);

   if (ReadDirectory(buffer, RDF_FILE, &list) IS ERR_Okay) {

      for (scan=list; scan; scan=scan->Next) {
         data = buffer;
         ((LONG *)data)[0] = CODE_ICON;
         data += sizeof(LONG);

         for (len=0; scan->Name[len]; len++);

         for (j=len; (j > 0) AND (scan->Name[j] != '.'); j--);
         if (scan->Name[j] IS '.') {
            len = j;
            scan->Name[j] = 0;
         }
         len++;
         len = (len + 3) & (~3);

         ((LONG *)data)[0] = sizeof(LONG) + sizeof(LONG) + len;

         data += sizeof(LONG);
         data += StrCopy(scan->Name, data, COPY_ALL);
         *data++ = 0; *data++ = 0; *data++ = 0; *data++ = 0;

         write.Buffer = buffer;
         write.Length = sizeof(LONG) + sizeof(LONG) + len;
         Action(AC_Write, Output, &write);
      }

      FreeMemory(list);
   }
}

/*****************************************************************************
** Internal: ResolveIconAssign()
**
** This function is called by the file system when it wants to resolve
** the icons: assignment.  It is also used to check for the existence of
** files.
**
** Icons can be accessed in the following formats:
**
**    icons:filename
**    icons:category/filename
**    icons:category/filename(size)
*/

static ERROR ResolveIconAssign(struct IconServer *Self, STRING Path, STRING Buffer, LONG BufferSize)
{
   struct Picture *picture, *resizepic;
   struct FileInfo cacheinfo, srcinfo;
   OBJECTPTR file, object;
   OBJECTID displayid;
   UBYTE category[60], icon[60];
   UBYTE *data;
   LONG width, height, type;
   WORD i, j, size, code;

   //DPrintF("IconServer:","ResolveAssign(%s)", Path);

   /*** Extract the category and the name of the icon ***/

   for (i=0; (Path[i] != ':'); i++) if (!Path[i]) return(ERR_Failed);
   i++;

   j = 0;
   while ((Path[i] != '/') AND (Path[i] != '\\') AND (Path[i] != '(') AND (Path[i])) category[j++] = Path[i++];
   category[j] = 0;

   if (Path[i]) {
      /*** Extract the icon name ***/
      while (Path[i]) i++;
      while ((i > 0) AND (Path[i-1] != '/') AND (Path[i-1] != '\\')) i--;
      j = 0;
      while ((Path[i] != '/') AND (Path[i] != '\\') AND (Path[i] != '(') AND (Path[i])) icon[j++] = Path[i++];
      icon[j] = 0;
   }
   else {
      /*** There is no listed category (the first string was a reference to the icon name) ***/
      for (j=0; category[j]; j++) icon[j] = category[j];
      icon[j] = 0;
      category[0] = 0;
   }

   /*** Strip the .png extension if one is specified ***/

   j -= 4;
   if ((j > 0) AND (StrCompare(".png", icon + j, 4, NULL) IS ERR_Okay)) {
      icon[j] = 0;
   }

   /*** Extract the icon size if one is specified ***/

   size = Self->FixedSize;
   if (Self->FixedSize < 16) {
      if (FastFindObject("SystemDisplay", NULL, &displayid, 1, NULL) IS ERR_Okay) {
         if (AccessObject(displayid, 3000, &object) IS ERR_Okay) {
            GetFields(object, FID_Width|TLONG,  &width,
                              FID_Height|TLONG, &height,
                              TAGEND);
            if (width < height) size = ((FLOAT)width) * Self->IconRatio / 100.0;
            else size = ((FLOAT)height) * Self->IconRatio / 100.0;

            ReleaseObject(object);
         }
      }
   }

   if (Path[i] IS '(') size = StrToInt(Path + i);
   if (size < 16) size = 16;
   else if (size > 256) size = 256;

   /* If the path ends with a trailing slash, then a directory has actually
   ** been requested and we can resolve the path to its absolute location.
   */

   while (Path[i]) i++;
   if ((Path[i-1] IS '/') OR (Path[i-1] IS '\\') OR (Path[i-1] IS ':')) {

      StrFormat(Buffer, BufferSize, "pictures:icons/%s/%s", Self->prvTheme, Path + 6);
      return(ERR_Okay);
   }

   /* Check the database to see if the requested icon exists at the specified
   ** category.  If it doesn't, clear the category so that we can scan for the
   ** icon in other areas.
   */

   if (category[0]) {
      data = Self->Database;
      while (data) {
         code = ((LONG *)data)[0];

         if (code IS CODE_END) {
            DPrintF("@IconServer:","Category \"%s\" does not exist.", category);
            category[0] = 0;
            break;
         }

         if (code IS CODE_CATEGORY) {
            if (StrCompare(category, data + (sizeof(LONG)*2), 0, STR_MATCHLENGTH) IS ERR_Okay) {
               /*** Scan within this category to see if the icon lives there ***/

               data = data + ((LONG *)data)[1];
               while (((LONG *)data)[0] IS CODE_ICON) {
                  if (StrCompare(icon, data + (sizeof(LONG)*2), 0, STR_MATCHLENGTH) IS ERR_Okay) break;
                  data = data + ((LONG *)data)[1];
               }

               if (((LONG *)data)[0] != CODE_ICON) {
                  DPrintF("@IconServer:","Icon \"%s\" does not exist at category \"%s\".", icon, category);
                  category[0] = 0;
               }
               break;
            }
         }

         data = data + ((LONG *)data)[1];
      }
   }

   /*** If no category was specified, scan the icon database for the icon ***/

   if (!category[0]) {
      DPrintF("IconServer:","Scanning database for a category for icon \"%s\".", icon);

      data = Self->Database;
      while (data) {
         code = ((LONG *)data)[0];

         /*** Return if we got to the end without any result ***/

         if (code IS CODE_END) return(ERR_Search);

         if (code IS CODE_CATEGORY) {
            StrCopy(data + (sizeof(LONG)*2), category, sizeof(category));
         }

         if (code IS CODE_ICON) {
            if (StrCompare(icon, data + (sizeof(LONG)*2), NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               /*** We have the correct category, so we just need to break the loop ***/
               break;
            }
         }

         data = data + ((LONG *)data)[1];
      }
   }

   /*** If the icon is requested at a small size, check for a tiny version ***/

   if (size IS 16) {
      StrFormat(Buffer, BufferSize, "pictures:icons/%s/%s/tiny/%s.png", Self->prvTheme, category, icon);
      if ((AnalyseLocation(Buffer, &type) IS ERR_Okay) AND (type IS LOC_FILE)) return(ERR_Okay);
   }

   /* Try to open the icon at its cache location.  If it is there, then we will use the
   ** cached version.  We also do some date checking here - if the date/time on the original
   ** is newer than the date on the cache, we'll need to regenerate the cached image.
   */

   StrFormat(Buffer, BufferSize, "user:temp/iconcache/%s/%s/%s%d.png", Self->prvTheme, category, icon, size);

   if (Self->VolatileIcons IS FALSE) {
      if (GetFileInfo(Buffer, &cacheinfo, sizeof(cacheinfo)) IS ERR_Okay) {
         StrFormat(Buffer, BufferSize, "pictures:icons/%s/%s/%s.png", Self->prvTheme, category, icon);
         if (GetFileInfo(Buffer, &srcinfo, sizeof(srcinfo)) IS ERR_Okay) {
            if (srcinfo.TimeStamp < cacheinfo.TimeStamp) {
               /*** Source has not been udpated, so we can use the existing cache file ***/
               StrFormat(Buffer, BufferSize, "user:temp/iconcache/%s/%s/%s%d.png", Self->prvTheme, category, icon, size);
               return(ERR_Okay);
            }
            else DPrintF("IconServer:","Regenerating new cache file for icon \"%s/%s\".", category, icon);
         }
      }
   }

   /* Load the original icon, resize it to the requested size and write the
   ** resulting picture file to the cache.  Then resolve the assignment so that
   ** it points to the cached file.
   */

   DPrintF("IconServer:","Resizing icon to new pixel size: %d", size);

   StrFormat(Buffer, BufferSize, "pictures:icons/%s/%s/%s", Self->prvTheme, category, icon);

   if (CreateObject(ID_PICTURE, NF_CHILD, &picture, NULL,
      FID_Location|TSTRING, Buffer,
      FID_Flags|TLONG,      PCF_32BITALPHA,
      TAGEND) IS ERR_Okay) {
      DOUBLE sizeratio;

      /*** Initialise the destination picture that we are going to use for resizing ***/

      if (NewObject(ID_PNG, NF_CHILD, &resizepic, NULL) != ERR_Okay) {
         acFree(picture);
         return(ERR_NewObject);
      }

      if (picture->Bitmap->Width > picture->Bitmap->Height) {
         sizeratio = (DOUBLE)size / (DOUBLE)picture->Bitmap->Width;
      }
      else sizeratio = (DOUBLE)size / (DOUBLE)picture->Bitmap->Height;

      resizepic->Flags |= PCF_NEW | (picture->Flags & (PCF_MASK|PCF_ALPHA|PCF_TRANSPARENT|PCF_32BITALPHA));
      resizepic->Bitmap->BitsPerPixel = picture->Bitmap->BitsPerPixel;
      resizepic->Bitmap->Width  = ((FLOAT)picture->Bitmap->Width)  * sizeratio;
      resizepic->Bitmap->Height = ((FLOAT)picture->Bitmap->Height) * sizeratio;

      if (acInit(resizepic) != ERR_Okay) {
         DPrintF("@IconServer:","Failed to initialise the resize picture space.");
         acFree(resizepic);
         acFree(picture);
         return(ERR_Init);
      }

      /*** Stretch the source into the destination ***/

      ActionTags(MT_BlitStretch, picture->Bitmap,
         resizepic->Bitmap, BSF_BILINEAR|BSF_FILTERSOURCE, 0, 0,
         picture->Bitmap->Width, picture->Bitmap->Height,
         0, 0, resizepic->Bitmap->Width, resizepic->Bitmap->Height);

      if ((picture->Mask) AND (resizepic->Mask)) {
         ActionTags(MT_BlitStretch, picture->Mask,
            resizepic->Mask, BSF_BILINEAR|BSF_FILTERSOURCE, -1, -1,
            picture->Mask->Width + 2, picture->Mask->Height + 2,
            0, 0, resizepic->Mask->Width, resizepic->Mask->Height);
      }

      /*** Save the new picture ***/

      StrFormat(Buffer, BufferSize, "user:temp/iconcache/%s/%s/%s%d.png", Self->prvTheme, category, icon, size);

      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING,  Buffer,
         FID_Flags|TLONG,       FL_NEW|FL_WRITE,
         FID_Permissions|TLONG, PERMIT_READ|PERMIT_WRITE,
         TAGEND) IS ERR_Okay) {

         ActionTags(AC_SaveToObject, resizepic, file->UniqueID);

         acFree(file);
      }

      acFree(resizepic);
      acFree(picture);
      return(ERR_Okay);
   }
   else {
      DPrintF("@IconServer:","Failed to open file \"%s\".", Buffer);
      return(ERR_CreateObject);
   }
}
